#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.algo.asha`."""

import hashlib

import numpy as np
import pytest

from orion.algo.asha import ASHA, Bracket, compute_budgets
from orion.algo.space import Fidelity, Integer, Real, Space


@pytest.fixture
def space():
    """Create a Space with a real dimension and a fidelity value."""
    space = Space()
    space.register(Real('lr', 'uniform', 0, 1))
    space.register(Fidelity('epoch', 1, 9, 3))
    return space


@pytest.fixture
def b_config(space):
    """Return a configuration for a bracket."""
    fidelity_dim = space.values()[0]
    num_rungs = 3
    budgets = np.logspace(
        np.log(fidelity_dim.low) / np.log(fidelity_dim.base),
        np.log(fidelity_dim.high) / np.log(fidelity_dim.base),
        num_rungs, base=fidelity_dim.base)
    return {'reduction_factor': fidelity_dim.base, 'budgets': budgets}


@pytest.fixture
def asha(b_config, space):
    """Return an instance of ASHA."""
    return ASHA(space)


@pytest.fixture
def bracket(b_config):
    """Return a `Bracket` instance configured with `b_config`."""
    return Bracket(None, b_config['reduction_factor'], b_config['budgets'])


@pytest.fixture
def rung_0():
    """Create fake points and objectives for rung 0."""
    points = np.linspace(0, 1, 9)
    return (1, {hashlib.md5(str([point]).encode('utf-8')).hexdigest():
            (point, (1, point)) for point in points})


@pytest.fixture
def rung_1(rung_0):
    """Create fake points and objectives for rung 1."""
    return (3, {hashlib.md5(str([value[0]]).encode('utf-8')).hexdigest(): value for value in
            map(lambda v: (v[0], (3, v[0])), sorted(rung_0[1].values()))})


@pytest.fixture
def rung_2(rung_1):
    """Create fake points and objectives for rung 1."""
    return (9, {hashlib.md5(str([value[0]]).encode('utf-8')).hexdigest(): value for value in
            map(lambda v: (v[0], (9, v[0])), sorted(rung_1[1].values()))})


def test_compute_budgets():
    """Verify proper computation of budgets on a logarithmic scale"""
    # Check typical values
    assert compute_budgets(1, 16, 4, 3) == [1, 4, 16]
    # Check rounding (max_resources is not a multiple of reduction_factor)
    assert compute_budgets(1, 30, 4, 3) == [1, 5, 30]
    # Check rounding (min_resources may be rounded below its actual value)
    assert compute_budgets(25, 1000, 2, 6) == [25, 52, 109, 229, 478, 1000]
    # Check min_resources
    assert compute_budgets(5, 125, 5, 3) == [5, 25, 125]
    # Check num_rungs
    assert compute_budgets(1, 16, 2, 5) == [1, 2, 4, 8, 16]


def test_compute_compressed_budgets():
    """Verify proper computation of budgets when scale is small and integer rounding creates
    duplicates
    """
    assert compute_budgets(1, 16, 2, 10) == [1, 2, 3, 4, 5, 6, 7, 9, 12, 16]

    with pytest.raises(ValueError) as exc:
        compute_budgets(1, 2, 2, 10)

    assert 'Cannot build budgets below max_resources' in str(exc.value)


class TestBracket():
    """Tests for the `Bracket` class."""

    def test_rungs_creation(self, bracket):
        """Test the creation of rungs for bracket 0."""
        assert len(bracket.rungs) == 3
        assert bracket.rungs[0][0] == 1
        assert bracket.rungs[1][0] == 3
        assert bracket.rungs[2][0] == 9

    def test_register(self, asha, bracket):
        """Check that a point is correctly registered inside a bracket."""
        bracket.asha = asha
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        bracket.register(point, 0.0)

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0][1]
        assert (0.0, point) == bracket.rungs[0][1][point_hash]

    def test_bad_register(self, asha, bracket):
        """Check that a non-valid point is not registered."""
        bracket.asha = asha

        with pytest.raises(IndexError) as ex:
            bracket.register((55, 0.0), 0.0)

        assert 'Bad fidelity level 55' in str(ex.value)

    def test_candidate_promotion(self, asha, bracket, rung_0):
        """Test that correct point is promoted."""
        bracket.asha = asha
        bracket.rungs[0] = rung_0

        point = bracket.get_candidate(0)

        assert point == (1, 0.0)

    def test_promotion_with_rung_1_hit(self, asha, bracket, rung_0):
        """Test that get_candidate gives us the next best thing if point is already in rung 1."""
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()
        bracket.asha = asha
        bracket.rungs[0] = rung_0
        bracket.rungs[1][1][point_hash] = (0.0, point)

        point = bracket.get_candidate(0)

        assert point == (1, 0.125)

    def test_no_promotion_when_rung_full(self, asha, bracket, rung_0, rung_1):
        """Test that get_candidate returns `None` if rung 1 is full."""
        bracket.asha = asha
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1

        point = bracket.get_candidate(0)

        assert point is None

    def test_no_promotion_if_not_enough_points(self, asha, bracket):
        """Test the get_candidate return None if there is not enough points ready."""
        bracket.asha = asha
        bracket.rungs[0] = (1, {hashlib.md5(str([0.0]).encode('utf-8')).hexdigest():
                                (0.0, (1, 0.0))})

        point = bracket.get_candidate(0)

        assert point is None

    def test_no_promotion_if_not_completed(self, asha, bracket, rung_0):
        """Test the get_candidate return None if trials are not completed."""
        bracket.asha = asha
        bracket.rungs[0] = rung_0
        rung = bracket.rungs[0][1]

        point = bracket.get_candidate(0)

        for p_id in rung.keys():
            rung[p_id] = (None, rung[p_id][1])

        point = bracket.get_candidate(0)

        assert point is None

    def test_is_done(self, bracket, rung_0):
        """Test that the `is_done` property works."""
        assert not bracket.is_done

        # Actual value of the point is not important here
        bracket.rungs[2] = (9, {'1': (1, 0.0)})

        assert bracket.is_done

    def test_update_rungs_return_candidate(self, asha, bracket, rung_1):
        """Check if a valid modified candidate is returned by update_rungs."""
        bracket.asha = asha
        bracket.rungs[1] = rung_1
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        candidate = bracket.update_rungs()

        assert point_hash in bracket.rungs[1][1]
        assert bracket.rungs[1][1][point_hash] == (0.0, (3, 0.0))
        assert candidate[0] == 9

    def test_update_rungs_return_no_candidate(self, asha, bracket, rung_1):
        """Check if no candidate is returned by update_rungs."""
        bracket.asha = asha

        candidate = bracket.update_rungs()

        assert candidate is None

    def test_repr(self, bracket, rung_0, rung_1, rung_2):
        """Test the string representation of Bracket"""
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1
        bracket.rungs[2] = rung_2

        assert str(bracket) == 'Bracket([1, 3, 9])'


class TestASHA():
    """Tests for the algo ASHA."""

    def test_register(self, asha, bracket, rung_0, rung_1):
        """Check that a point is registered inside the bracket."""
        asha.brackets = [bracket]
        bracket.asha = asha
        bracket.rungs = [rung_0, rung_1]
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        asha.observe([point], [{'objective': 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0][1]
        assert (0.0, point) == bracket.rungs[0][1][point_hash]

    def test_register_bracket_multi_fidelity(self, space, b_config):
        """Check that a point is registered inside the same bracket for diff fidelity."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 1
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        asha.observe([point], [{'objective': 0.0}])

        bracket = asha.brackets[0]

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0][1]
        assert (0.0, point) == bracket.rungs[0][1][point_hash]

        fidelity = 3
        point = [fidelity, value]
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        asha.observe([point], [{'objective': 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[1][1]
        assert (0.0, point) != bracket.rungs[0][1][point_hash]
        assert (0.0, point) == bracket.rungs[1][1][point_hash]

    def test_register_next_bracket(self, space, b_config):
        """Check that a point is registered inside the good bracket when higher fidelity."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 3
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        asha.observe([point], [{'objective': 0.0}])

        assert sum(len(rung[1]) for rung in asha.brackets[0].rungs) == 0
        assert sum(len(rung[1]) for rung in asha.brackets[1].rungs) == 1
        assert sum(len(rung[1]) for rung in asha.brackets[2].rungs) == 0
        assert point_hash in asha.brackets[1].rungs[0][1]
        assert (0.0, point) == asha.brackets[1].rungs[0][1][point_hash]

        value = 51
        fidelity = 9
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode('utf-8')).hexdigest()

        asha.observe([point], [{'objective': 0.0}])

        assert sum(len(rung[1]) for rung in asha.brackets[0].rungs) == 0
        assert sum(len(rung[1]) for rung in asha.brackets[1].rungs) == 1
        assert sum(len(rung[1]) for rung in asha.brackets[2].rungs) == 1
        assert point_hash in asha.brackets[2].rungs[0][1]
        assert (0.0, point) == asha.brackets[2].rungs[0][1][point_hash]

    def test_register_invalid_fidelity(self, space, b_config):
        """Check that a point cannot registered if fidelity is invalid."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 2
        point = (fidelity, value)

        with pytest.raises(ValueError) as ex:
            asha.observe([point], [{'objective': 0.0}])

        assert 'No bracket found for point' in str(ex.value)

    def test_register_corrupted_db(self, caplog, space, b_config):
        """Check that a point cannot registered if passed in order diff than fidelity."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 3
        point = (fidelity, value)

        asha.observe([point], [{'objective': 0.0}])
        assert 'Point registered to wrong bracket' not in caplog.text

        fidelity = 1
        point = [fidelity, value]

        caplog.clear()
        asha.observe([point], [{'objective': 0.0}])
        assert 'Point registered to wrong bracket' in caplog.text

    def test_get_id(self, space, b_config):
        """Test valid id of points"""
        asha = ASHA(space, num_brackets=3)

        assert asha.get_id(['whatever', 1]) == asha.get_id(['is here', 1])
        assert asha.get_id(['whatever', 1]) != asha.get_id(['is here', 2])

    def test_get_id_multidim(self, b_config):
        """Test valid id for points with dim of shape > 1"""
        space = Space()
        space.register(Fidelity('epoch', 1, 9, 3))
        space.register(Real('lr', 'uniform', 0, 1, shape=2))

        asha = ASHA(space, num_brackets=3)

        assert asha.get_id(['whatever', [1, 1]]) == asha.get_id(['is here', [1, 1]])
        assert asha.get_id(['whatever', [1, 1]]) != asha.get_id(['is here', [2, 2]])

    def test_suggest_new(self, monkeypatch, asha, bracket, rung_0, rung_1, rung_2):
        """Test that a new point is sampled."""
        asha.brackets = [bracket]
        bracket.asha = asha

        def sample(num=1, seed=None):
            return [('fidelity', 0.5)]

        monkeypatch.setattr(asha.space, 'sample', sample)

        points = asha.suggest()

        assert points == [(1, 0.5)]

    def test_suggest_duplicates(self, monkeypatch, asha, bracket, rung_0, rung_1, rung_2):
        """Test that sampling collisions are handled."""
        asha.brackets = [bracket]
        bracket.asha = asha

        duplicate_point = ('fidelity', 0.0)
        new_point = ('fidelity', 0.5)

        duplicate_id = hashlib.md5(str([duplicate_point]).encode('utf-8')).hexdigest()
        bracket.rungs[0] = (1, {duplicate_id: (0.0, duplicate_point)})

        asha.trial_info[asha.get_id(duplicate_point)] = bracket

        points = [duplicate_point, new_point]

        def sample(num=1, seed=None):
            return [points.pop(0)]

        monkeypatch.setattr(asha.space, 'sample', sample)

        assert asha.suggest()[0][1] == new_point[1]
        assert len(points) == 0

    def test_suggest_inf_duplicates(self, monkeypatch, asha, bracket, rung_0, rung_1, rung_2):
        """Test that sampling inf collisions raises runtime error."""
        asha.brackets = [bracket]
        bracket.asha = asha

        zhe_point = ('fidelity', 0.0)
        asha.trial_info[asha.get_id(zhe_point)] = bracket

        def sample(num=1, seed=None):
            return [zhe_point]

        monkeypatch.setattr(asha.space, 'sample', sample)

        with pytest.raises(RuntimeError) as exc:
            asha.suggest()

        assert 'ASHA keeps sampling already existing points.' in str(exc.value)

    def test_suggest_in_finite_cardinality(self):
        """Test that suggest None when search space is empty"""
        space = Space()
        space.register(Integer('yolo1', 'uniform', 0, 6))
        space.register(Fidelity('epoch', 1, 9, 3))

        asha = ASHA(space)
        for i in range(6):
            asha.observe([(1, i)], [{'objective': i}])

        for i in range(2):
            asha.observe([(3, i)], [{'objective': i}])

        assert asha.suggest() is None

    def test_suggest_promote(self, asha, bracket, rung_0):
        """Test that correct point is promoted and returned."""
        asha.brackets = [bracket]
        bracket.asha = asha
        bracket.rungs[0] = rung_0

        points = asha.suggest()

        assert points == [(3, 0.0)]

    def test_suggest_opt_out(self, asha, bracket, rung_0, rung_1, rung_2):
        """Test that ASHA opts out when last rung is full."""
        asha.brackets = [bracket]
        bracket.asha = asha
        bracket.rungs[1] = rung_1
        bracket.rungs[2] = rung_2

        points = asha.suggest()

        assert points is None

    def test_seed_rng(self, asha):
        """Test that algo is seeded properly"""
        asha.seed_rng(1)
        a = asha.suggest(1)[0]
        assert not np.allclose(a, asha.suggest(1)[0])

        asha.seed_rng(1)
        assert np.allclose(a, asha.suggest(1)[0])

    def test_set_state(self, asha):
        """Test that state is reset properly"""
        asha.seed_rng(1)
        state = asha.state_dict
        point = asha.suggest(1)[0]
        assert not np.allclose(point, asha.suggest(1)[0])

        asha.set_state(state)
        assert point == asha.suggest(1)[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.algo.tpe`."""

import numpy
import pytest
from scipy.stats import norm

from orion.algo.space import Categorical, Fidelity, Integer, Real, Space
from orion.algo.tpe import adaptive_parzen_estimator, CategoricalSampler, \
    compute_max_ei_point, GMMSampler, ramp_up_weights, TPE


@pytest.fixture()
def space():
    """Return an optimization space"""
    space = Space()

    dim1 = Real('yolo1', 'uniform', -10, 20)
    space.register(dim1)

    dim2 = Integer('yolo2', 'uniform', -5, 10)
    space.register(dim2)

    categories = ['a', 0.1, 2, 'c']
    dim3 = Categorical('yolo3', categories)
    space.register(dim3)

    return space


@pytest.fixture
def tpe(space):
    """Return an instance of TPE."""
    return TPE(space, seed=1)


def test_compute_max_ei_point():
    """Test that max ei point is computed correctly"""
    points = numpy.linspace(-3, 3, num=10)
    below_likelis = numpy.linspace(0.5, 0.9, num=10)
    above_likes = numpy.linspace(0.2, 0.5, num=10)

    numpy.random.shuffle(below_likelis)
    numpy.random.shuffle(above_likes)
    max_ei_index = (below_likelis - above_likes).argmax()

    max_ei_point = compute_max_ei_point(points, below_likelis, above_likes)
    assert max_ei_point == points[max_ei_index]


def test_ramp_up_weights():
    """Test TPE adjust observed points correctly"""
    weights = ramp_up_weights(25, 15, True)
    assert len(weights) == 25
    assert numpy.all(weights == 1.0)

    weights = ramp_up_weights(25, 15, False)
    assert len(weights) == 25
    assert numpy.all(weights[:10] == (numpy.linspace(1.0 / 25, 1.0, num=10)))
    assert numpy.all(weights[10:] == 1.0)

    weights = ramp_up_weights(10, 15, False)
    assert len(weights) == 10
    assert numpy.all(weights == 1.0)

    weights = ramp_up_weights(25, 0, False)
    assert len(weights) == 25
    assert numpy.all(weights == (numpy.linspace(1.0 / 25, 1.0, num=25)))


def test_adaptive_parzen_normal_estimator():
    """Test adaptive parzen estimator"""
    low = -1
    high = 5

    obs_mus = [1.2]
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=25)
    assert list(mus) == [1.2, 2]
    assert list(sigmas) == [3, 6]
    assert list(weights) == [1.0 / 2, 1.0 / 2]

    obs_mus = [3.4]
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=0.5,
                                                     equal_weight=False, flat_num=25)
    assert list(mus) == [2, 3.4]
    assert list(sigmas) == [6, 3]
    assert list(weights) == [0.5 / 1.5, 1.0 / 1.5]

    obs_mus = numpy.linspace(-1, 5, num=30, endpoint=False)
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=25)

    ramp = numpy.linspace(1.0 / 30, 1.0, num=30 - 25)
    full = numpy.ones(25 + 1)
    all_weights = (numpy.concatenate([ramp, full]))

    assert len(mus) == len(sigmas) == len(weights) == 30 + 1
    assert numpy.all(weights[:30 - 25] == ramp / all_weights.sum())
    assert numpy.all(weights[30 - 25:] == 1 / all_weights.sum())
    assert numpy.all(sigmas == 6 / 10)


def test_adaptive_parzen_normal_estimator_weight():
    """Test the weight for the normal components"""
    obs_mus = numpy.linspace(-1, 5, num=30, endpoint=False)
    low = -1
    high = 5

    # equal weight
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=True, flat_num=25)
    assert numpy.all(weights == 1 / 31)
    assert numpy.all(sigmas == 6 / 10)

    # prior weight
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=0.5,
                                                     equal_weight=False, flat_num=25)

    ramp = numpy.linspace(1.0 / 30, 1.0, num=30 - 25)
    full = numpy.ones(25 + 1)
    all_weights = (numpy.concatenate([ramp, full]))
    prior_pos = numpy.searchsorted(mus, 2)
    all_weights[prior_pos] = 0.5

    assert numpy.all(weights[:30 - 25] == (numpy.linspace(1.0 / 30, 1.0, num=30 - 25) /
                                           all_weights.sum()))
    assert numpy.all(weights[33 - 25:prior_pos] == 1 / all_weights.sum())
    assert weights[prior_pos] == 0.5 / all_weights.sum()
    assert numpy.all(weights[prior_pos + 1:] == 1 / all_weights.sum())
    assert numpy.all(sigmas == 6 / 10)

    # full weights number
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=15)

    ramp = numpy.linspace(1.0 / 30, 1.0, num=30 - 15)
    full = numpy.ones(15 + 1)
    all_weights = (numpy.concatenate([ramp, full]))
    prior_pos = numpy.searchsorted(mus, 2)
    all_weights[prior_pos] = 1.0

    assert numpy.all(weights[:30 - 15] == (numpy.linspace(1.0 / 30, 1.0, num=30 - 15) /
                                           all_weights.sum()))
    assert numpy.all(weights[30 - 15:] == 1 / all_weights.sum())
    assert numpy.all(sigmas == 6 / 10)


def test_adaptive_parzen_normal_estimator_sigma_clip():
    """Test that the magic clip of sigmas for parzen estimator"""
    low = -1
    high = 5

    obs_mus = numpy.linspace(-1, 5, num=8, endpoint=False)
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=25)
    assert len(mus) == len(sigmas) == len(weights) == 8 + 1
    assert numpy.all(weights == 1 / 9)
    assert numpy.all(sigmas == 6 / 8)

    obs_mus = numpy.random.uniform(-1, 5, 30)
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=25)

    assert len(mus) == len(sigmas) == len(weights) == 30 + 1
    assert numpy.all(weights[-25:] == weights[-1])
    assert numpy.all(sigmas <= 6) and numpy.all(sigmas >= 6 / 10)

    obs_mus = numpy.random.uniform(-1, 5, 400)
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=25)

    assert len(mus) == len(sigmas) == len(weights) == 400 + 1
    assert numpy.all(weights[-25:] == weights[-1])
    assert numpy.all(sigmas <= 6) and numpy.all(sigmas >= 6 / 20)

    obs_mus = numpy.random.uniform(-1, 5, 10000)
    mus, sigmas, weights = adaptive_parzen_estimator(obs_mus, low, high, prior_weight=1.0,
                                                     equal_weight=False, flat_num=25)

    assert len(mus) == len(sigmas) == len(weights) == 10000 + 1
    assert numpy.all(weights[-25:] == weights[-1])
    assert numpy.all(sigmas <= 6) and numpy.all(sigmas >= 6 / 100)


class TestCategoricalSampler():
    """Tests for TPE Categorical Sampler"""

    def test_cat_sampler_creation(self, tpe):
        """Test CategoricalSampler creation"""
        obs = [0, 3, 9]
        choices = list(range(-5, 5))
        cat_sampler = CategoricalSampler(tpe, obs, choices)
        assert len(cat_sampler.weights) == len(choices)

        obs = [0, 3, 9]
        choices = ['a', 'b', 11, 15, 17, 18, 19, 20, 25, 'c']
        cat_sampler = CategoricalSampler(tpe, obs, choices)

        assert len(cat_sampler.weights) == len(choices)

        tpe.equal_weight = True
        tpe.prior_weight = 1.0
        obs = numpy.random.randint(0, 10, 100)
        cat_sampler = CategoricalSampler(tpe, obs, choices)
        counts_obs = numpy.bincount(obs) + 1.0
        weights = counts_obs / counts_obs.sum()

        assert numpy.all(cat_sampler.weights == weights)

        tpe.equal_weight = False
        tpe.prior_weight = 0.5
        tpe.full_weight_num = 30
        obs = numpy.random.randint(0, 10, 100)

        cat_sampler = CategoricalSampler(tpe, obs, choices)

        ramp = numpy.linspace(1.0 / 100, 1.0, num=100 - 30)
        full = numpy.ones(30)
        ramp_weights = (numpy.concatenate([ramp, full]))

        counts_obs = numpy.bincount(obs, weights=ramp_weights) + 0.5
        weights = counts_obs / counts_obs.sum()

        assert numpy.all(cat_sampler.weights == weights)

    def test_sample(self, tpe):
        """Test CategoricalSampler sample function"""
        obs = numpy.random.randint(0, 10, 100)
        choices = ['a', 'b', 11, 15, 17, 18, 19, 20, 25, 'c']
        cat_sampler = CategoricalSampler(tpe, obs, choices)

        points = cat_sampler.sample(25)

        assert len(points) == 25
        assert numpy.all(points >= 0)
        assert numpy.all(points < 10)

        weights = numpy.linspace(1, 10, num=10) ** 3
        numpy.random.shuffle(weights)
        weights = weights / weights.sum()
        cat_sampler = CategoricalSampler(tpe, obs, choices)
        cat_sampler.weights = weights

        points = cat_sampler.sample(10000)
        points = numpy.array(points)
        hist = numpy.bincount(points)

        assert numpy.all(hist.argsort() == weights.argsort())
        assert len(points) == 10000
        assert numpy.all(points >= 0)
        assert numpy.all(points < 10)

    def test_get_loglikelis(self, tpe):
        """Test to get log likelis of points"""
        obs = numpy.random.randint(0, 10, 100)
        choices = ['a', 'b', 11, 15, 17, 18, 19, 20, 25, 'c']
        cat_sampler = CategoricalSampler(tpe, obs, choices)

        points = cat_sampler.sample(25)

        likelis = cat_sampler.get_loglikelis(points)

        assert numpy.all(likelis == numpy.log(numpy.asarray(cat_sampler.weights)[points]))


class TestGMMSampler():
    """Tests for TPE GMM Sampler"""

    def test_gmm_sampler_creation(self, tpe):
        """Test GMMSampler creation"""
        mus = numpy.linspace(-3, 3, num=12, endpoint=False)
        sigmas = [0.5] * 12

        gmm_sampler = GMMSampler(tpe, mus, sigmas, -3, 3)

        assert len(gmm_sampler.weights) == 12
        assert len(gmm_sampler.pdfs) == 12

    def test_sample(self, tpe):
        """Test GMMSampler sample function"""
        mus = numpy.linspace(-3, 3, num=12, endpoint=False)
        sigmas = [0.5] * 12

        gmm_sampler = GMMSampler(tpe, mus, sigmas, -3, 3)
        points = gmm_sampler.sample(25)
        points = numpy.array(points)

        assert len(points) <= 25
        assert numpy.all(points >= -3)
        assert numpy.all(points < 3)

        mus = numpy.linspace(-10, 10, num=10, endpoint=False)
        sigmas = [0.00001] * 10
        weights = numpy.linspace(1, 10, num=10) ** 3
        numpy.random.shuffle(weights)
        weights = weights / weights.sum()

        gmm_sampler = GMMSampler(tpe, mus, sigmas, -11, 9, weights)
        points = gmm_sampler.sample(10000)
        points = numpy.array(points)
        hist = numpy.histogram(points, bins=[-11, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9])

        assert numpy.all(hist[0].argsort() == numpy.array(weights).argsort())
        assert numpy.all(points >= -11)
        assert numpy.all(points < 9)

    def test_get_loglikelis(self):
        """Test to get log likelis of points"""
        mus = numpy.linspace(-10, 10, num=10, endpoint=False)
        weights = numpy.linspace(1, 10, num=10) ** 3
        numpy.random.shuffle(weights)
        weights = weights / weights.sum()

        sigmas = [0.00001] * 10
        gmm_sampler = GMMSampler(tpe, mus, sigmas, -11, 9, weights)

        points = [mus[7]]
        pdf = norm(mus[7], sigmas[7])
        point_likeli = numpy.log(pdf.pdf(mus[7]) * weights[7])
        likelis = gmm_sampler.get_loglikelis(points)

        assert list(likelis) == point_likeli
        assert likelis[0] == point_likeli

        sigmas = [2] * 10
        gmm_sampler = GMMSampler(tpe, mus, sigmas, -11, 9, weights)

        log_pdf = []
        pdfs = []
        for i in range(10):
            pdfs.append(norm(mus[i], sigmas[i]))
        for pdf, weight in zip(pdfs, weights):
            log_pdf.append(numpy.log(pdf.pdf(0) * weight))
        point_likeli = numpy.log(numpy.sum(numpy.exp(log_pdf)))

        points = numpy.random.uniform(-11, 9, 30)
        points = numpy.insert(points, 10, 0)
        likelis = gmm_sampler.get_loglikelis(points)

        point_likeli = numpy.format_float_scientific(point_likeli, precision=10)
        gmm_likeli = numpy.format_float_scientific(likelis[10], precision=10)
        assert point_likeli == gmm_likeli
        assert len(likelis) == len(points)


class TestTPE():
    """Tests for the algo TPE."""

    def test_seed_rng(self, tpe):
        """Test that algo is seeded properly"""
        tpe.seed_rng(1)
        a = tpe.suggest(1)[0]
        assert not numpy.allclose(a, tpe.suggest(1)[0])

        tpe.seed_rng(1)
        assert numpy.allclose(a, tpe.suggest(1)[0])

    def test_set_state(self, tpe):
        """Test that state is reset properly"""
        tpe.seed_rng(1)
        state = tpe.state_dict
        a = tpe.suggest(1)[0]
        assert not numpy.allclose(a, tpe.suggest(1)[0])

        tpe.set_state(state)
        assert numpy.allclose(a, tpe.suggest(1)[0])

    def test_unsupported_space(self):
        """Test tpe only work for supported search space"""
        space = Space()
        dim1 = Real('yolo1', 'uniform', -10, 10)
        space.register(dim1)
        dim2 = Real('yolo2', 'reciprocal', 10, 20)
        space.register(dim2)
        categories = ['a', 0.1, 2, 'c']
        dim3 = Categorical('yolo3', categories)
        space.register(dim3)
        dim4 = Fidelity('epoch', 1, 9, 3)
        space.register(dim4)
        TPE(space)

        space = Space()
        dim = Real('yolo1', 'norm', 0.9)
        space.register(dim)

        with pytest.raises(ValueError) as ex:
            TPE(space)

        assert 'TPE now only supports uniform, loguniform, uniform discrete and choices' \
               in str(ex.value)

        space = Space()
        dim = Real('yolo1', 'uniform', 0.9, shape=(2, 1))
        space.register(dim)

        with pytest.raises(ValueError) as ex:
            TPE(space)

        assert 'TPE now only supports 1D shape' in str(ex.value)

    def test_split_trials(self, tpe):
        """Test observed trials can be split based on TPE gamma"""
        space = Space()
        dim1 = Real('yolo1', 'uniform', -3, 6)
        space.register(dim1)

        tpe.space = space

        points = numpy.linspace(-3, 3, num=10, endpoint=False)
        results = numpy.linspace(0, 1, num=10, endpoint=False)
        points_results = list(zip(points, results))
        numpy.random.shuffle(points_results)
        points, results = zip(*points_results)
        for point, result in zip(points, results):
            tpe.observe([[point]], [{'objective': result}])

        tpe.gamma = 0.25
        below_points, above_points = tpe.split_trials()

        assert below_points == [[-3.0], [-2.4], [-1.8]]
        assert len(above_points) == 7

        tpe.gamma = 0.2
        below_points, above_points = tpe.split_trials()

        assert below_points == [[-3.0], [-2.4]]
        assert len(above_points) == 8

    def test_sample_int_dimension(self):
        """Test sample values for a integer dimension"""
        space = Space()
        dim1 = Integer('yolo1', 'uniform', -10, 20)
        space.register(dim1)

        dim2 = Integer('yolo2', 'uniform', -5, 10, shape=(2))
        space.register(dim2)

        tpe = TPE(space)

        obs_points = numpy.random.randint(-10, 10, 100)
        below_points = [obs_points[:25]]
        above_points = [obs_points[25:]]
        points = tpe.sample_one_dimension(dim1, 1,
                                          below_points, above_points, tpe._sample_int_point)
        points = numpy.asarray(points)
        assert len(points) == 1
        assert all(points >= -10)
        assert all(points < 10)

        obs_points_below = numpy.random.randint(-10, 0, 25).reshape(1, 25)
        obs_points_above = numpy.random.randint(0, 10, 75).reshape(1, 75)
        points = tpe.sample_one_dimension(dim1, 1,
                                          obs_points_below, obs_points_above, tpe._sample_int_point)
        points = numpy.asarray(points)
        assert len(points) == 1
        assert all(points >= -10)
        assert all(points < 0)

        obs_points = numpy.random.randint(-5, 5, 100)
        below_points = [obs_points[:25], obs_points[25:50]]
        above_points = [obs_points[50:75], obs_points[75:]]
        points = tpe.sample_one_dimension(dim2, 2,
                                          below_points, above_points, tpe._sample_int_point)
        points = numpy.asarray(points)
        assert len(points) == 2
        assert all(points >= -10)
        assert all(points < 10)

        tpe.n_ei_candidates = 0
        points = tpe.sample_one_dimension(dim2, 2,
                                          below_points, above_points, tpe._sample_int_point)
        assert len(points) == 0

    def test_sample_categorical_dimension(self):
        """Test sample values for a categorical dimension"""
        space = Space()
        categories = ['a', 'b', 11, 15, 17, 18, 19, 20, 25, 'c']
        dim1 = Categorical('yolo1', categories)
        space.register(dim1)
        dim2 = Categorical('yolo2', categories, shape=(2))
        space.register(dim2)

        tpe = TPE(space)

        obs_points = numpy.random.randint(0, 10, 100)
        obs_points = [categories[point] for point in obs_points]
        below_points = [obs_points[:25]]
        above_points = [obs_points[25:]]
        points = tpe.sample_one_dimension(dim1, 1,
                                          below_points, above_points, tpe._sample_categorical_point)
        assert len(points) == 1
        assert points[0] in categories

        obs_points_below = numpy.random.randint(0, 3, 25)
        obs_points_above = numpy.random.randint(3, 10, 75)
        below_points = [[categories[point] for point in obs_points_below]]
        above_points = [[categories[point] for point in obs_points_above]]
        points = tpe.sample_one_dimension(dim1, 1,
                                          below_points, above_points, tpe._sample_categorical_point)
        assert len(points) == 1
        assert points[0] in categories[:3]

        obs_points = numpy.random.randint(0, 10, 100)
        obs_points = [categories[point] for point in obs_points]
        below_points = [obs_points[:25], obs_points[25:50]]
        above_points = [obs_points[50:75], obs_points[75:]]

        points = tpe.sample_one_dimension(dim2, 2,
                                          below_points, above_points, tpe._sample_categorical_point)
        assert len(points) == 2
        assert points[0] in categories
        assert points[1] in categories

        tpe.n_ei_candidates = 0
        points = tpe.sample_one_dimension(dim2, 2,
                                          below_points, above_points, tpe._sample_categorical_point)
        assert len(points) == 0

    def test_sample_real_dimension(self):
        """Test sample values for a real dimension"""
        space = Space()
        dim1 = Real('yolo1', 'uniform', -10, 20)
        space.register(dim1)
        dim2 = Real('yolo2', 'uniform', -5, 10, shape=(2))
        space.register(dim2)
        dim3 = Real('yolo3', 'reciprocal', 1, 20)
        space.register(dim3)

        tpe = TPE(space)
        points = numpy.random.uniform(-10, 10, 20)
        below_points = [points[:8]]
        above_points = [points[8:]]
        points = tpe._sample_real_dimension(dim1, 1,
                                            below_points, above_points)
        points = numpy.asarray(points)
        assert len(points) == 1
        assert all(points >= -10)
        assert all(points < 10)

        points = numpy.random.uniform(1, 20, 20)
        below_points = [points[:8]]
        above_points = [points[8:]]
        points = tpe._sample_real_dimension(dim3, 1,
                                            below_points, above_points)
        points = numpy.asarray(points)
        assert len(points) == 1
        assert all(points >= 1)
        assert all(points < 20)

        below_points = numpy.random.uniform(-10, 0, 25).reshape(1, 25)
        above_points = numpy.random.uniform(0, 10, 75).reshape(1, 75)
        points = tpe._sample_real_dimension(dim1, 1,
                                            below_points, above_points)
        points = numpy.asarray(points)
        assert len(points) == 1
        assert all(points >= -10)
        assert all(points < 0)

        points = numpy.random.uniform(-5, 5, 32)
        below_points = [points[:8], points[8:16]]
        above_points = [points[16:24], points[24:]]
        points = tpe._sample_real_dimension(dim2, 2,
                                            below_points, above_points)
        points = numpy.asarray(points)
        assert len(points) == 2
        assert all(points >= -10)
        assert all(points < 10)

        tpe.n_ei_candidates = 0
        points = tpe._sample_real_dimension(dim2, 2,
                                            below_points, above_points)
        assert len(points) == 0

    def test_suggest(self, tpe):
        """Test suggest with no shape dimensions"""
        tpe.n_initial_points = 10
        results = numpy.random.random(10)
        for i in range(10):
            point = tpe.suggest(1)
            assert len(point) == 1
            assert len(point[0]) == 3
            assert not isinstance(point[0][0], tuple)
            tpe.observe(point, [{'objective': results[i]}])

        point = tpe.suggest(1)
        assert len(point) == 1
        assert len(point[0]) == 3
        assert not isinstance(point[0][0], tuple)

    def test_1d_shape(self, tpe):
        """Test suggest with 1D shape dimensions"""
        space = Space()
        dim1 = Real('yolo1', 'uniform', -3, 6, shape=(2))
        space.register(dim1)
        dim2 = Real('yolo2', 'uniform', -2, 4)
        space.register(dim2)

        tpe.space = space

        tpe.n_initial_points = 10
        results = numpy.random.random(10)
        for i in range(10):
            point = tpe.suggest(1)
            assert len(point) == 1
            assert len(point[0]) == 2
            assert len(point[0][0]) == 2
            tpe.observe(point, [{'objective': results[i]}])

        point = tpe.suggest(1)
        assert len(point) == 1
        assert len(point[0]) == 2
        assert len(point[0][0]) == 2

    def test_suggest_initial_points(self, tpe, monkeypatch):
        """Test that initial points can be sampled correctly"""
        points = [(i, i - 6, 'c') for i in range(1, 12)]

        global index
        index = 0

        def sample(num=1, seed=None):
            global index
            pts = points[index:index + num]
            index += num
            return pts

        monkeypatch.setattr(tpe.space, 'sample', sample)

        tpe.n_initial_points = 10
        results = numpy.random.random(10)
        for i in range(1, 11):
            point = tpe.suggest(1)[0]
            assert point == (i, i - 6, 'c')
            tpe.observe([point], [{'objective': results[i - 1]}])

        point = tpe.suggest(1)[0]
        assert point != (11, 5, 'c')

    def test_suggest_ei_candidates(self, tpe):
        """Test suggest with no shape dimensions"""
        tpe.n_initial_points = 2
        tpe.n_ei_candidates = 0

        results = numpy.random.random(2)
        for i in range(2):
            point = tpe.suggest(1)
            assert len(point) == 1
            assert len(point[0]) == 3
            assert not isinstance(point[0][0], tuple)
            tpe.observe(point, [{'objective': results[i]}])

        point = tpe.suggest(1)
        assert not point

        tpe.n_ei_candidates = 24
        point = tpe.suggest(1)
        assert len(point) > 0



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
"""
:mod:`orion.core.evc.adapters` -- Adapters to connect experiments within the EVC system
=======================================================================================

.. module:: adapters
   :platform: Unix
   :synopsis: Adapters to connect experiments within the EVC system

Experiment branches because of changes in the configuration or the user's code. This make
experiments incompatible with one another unless we define adapters such that a branched experiment
B can access trials from parent experiment A.

Adapters have two main methods, forward and backward. The method forward defines how trials from
parent experiment A are filtered or adapted to be compatible with experiment B. Modifications are
only applied at execution time and are not saved anywhere in the database. Adapters only provides a
view on an experiment.

There is adapters for

    * Dimension addition
    * Dimension deletion
    * Dimension renaming
    * Change of dimension prior
    * Change of algorithm
    * Change of code
    * Combining different adapters

Adapters all have a `to_dict` method which provides sufficient information to rebuild the adapter.
This is to facilitate save of adapters in a database and retrieval.

Adapters can be build using the factory class `Adapter(**kwargs)` or using
`Adapter.build(list_of_dicts)`.

"""
from abc import (ABCMeta, abstractmethod)
import copy

from orion.core.io.space_builder import DimensionBuilder
from orion.core.utils import Factory
from orion.core.worker.trial import Trial


class BaseAdapter(object, metaclass=ABCMeta):
    """Base class describing what an adapter can do."""

    @abstractmethod
    def forward(self, trials):
        """Adapt trials of the parent experiment such that they are compatible to the child
        experiment

        Parameters
        ----------
        trials: list of :class:`orion.core.worker.trial.Trial`
            List of :class:`orion.core.worker.trial.Trial` coming from the parent experiment

        """
        pass

    @abstractmethod
    def backward(self, trials):
        """Adapt trials of the child experiment such that they are compatible to the parent
        experiment

        Parameters
        ----------
        trials: list of :class:`orion.core.worker.trial.Trial`
            List of :class:`orion.core.worker.trial.Trial` coming from the child experiment

        """
        pass

    @abstractmethod
    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        This method is intended for single adapters only.
        For coherence, method configuration is used by all adapters, which is a list of dictionary
        configurations.

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.configuration`
        """
        pass

    @property
    def configuration(self):
        """Provide the configuration of the adapter.

        For simplicity, the configuration is always returned as if the adapter is a
        CompositeAdapter, therefore no matter if the adapter is composite or not it will return a
        list of configurations.

        The dictionaries of the list can be used to recreate adapters, for any adapter class.

        Examples
        --------
        .. code-block:: python
           :linenos:

            configuration = a_dummy_adapter.configuration[0]
            another_dummy_adapter = Adapter.build([configuration])
            assert another_dummy_adapter.configuration[0] == configuration

        Returns
        -------
        dict
            Configuration as a dictionary

        """
        return [self.to_dict()]


class CompositeAdapter(BaseAdapter):
    """Adapter which group many other adapters needed to connect two experiments

    Attributes
    ----------
    adapters: instances of `orion.core.evc.adapters.BaseAdapter`
        List of adaptors which are applied sequentially

    """

    def __init__(self, *adapters):
        """Initialize with adapters

        Parameters
        ----------
        adapters: instances of `orion.core.evc.adapters.BaseAdapter`
            List of adaptors which are applied sequentially

        """
        if any(not isinstance(adapter, BaseAdapter) for adapter in adapters):
            wrong_object_type = [type(adapter) for adapter in adapters
                                 if not isinstance(adapter, BaseAdapter)][0]
            raise TypeError("Provided adapters must be adapter objects, not '%s'" %
                            str(wrong_object_type))

        self.adapters = adapters

    def forward(self, trials):
        """Apply the adaptors on the parent's trials

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        for adapter in self.adapters:
            trials = adapter.forward(trials)

        return trials

    def backward(self, trials):
        """Apply the adaptors backward and in reverse order on the parent's trials

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        for adapter in self.adapters[::-1]:
            trials = adapter.backward(trials)

        return trials

    def to_dict(self):
        """Return nothing since it is not valid for CompositeAdapter

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
            :meth:`orion.core.evc.adapters.CompositeAdapter.configuration`
        """
        pass

    @property
    def configuration(self):
        """Provide the configuration of the adapter.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.configuration`
        """
        if len(self.adapters) > 1:
            return [adapter.configuration
                    if len(adapter.configuration) > 1
                    else adapter.configuration[0]
                    for adapter in self.adapters]
        elif self.adapters:
            return self.adapters[0].configuration

        return []


def apply_if_valid(name, trial, callback=None, raise_if_not=True):
    """Detect a parameter in trial and call a callback on it if provided

    Parameters
    ----------
    name: str
        Name of the param to look for
    trial: `orion.core.worker.trial.Trial`
        Instance of trial to investigate
    callback: None of callable
        Function to call with (trial, param) with a parameter is found.
        Defaults to None
    raise_if_not: bool
        raises RuntimeError if no parameter is found.
        Defaults to True.

    Returns
    -------
    bool
        False if parameter is not found and `raise_if_not is False`.
        True if parameter is found and callback is None.
        Else, output of callback(trial, item).

    """
    for param in trial._params:  # pylint: disable=protected-access
        if param.name == name:
            return callback is None or callback(trial, param)

    if raise_if_not:
        raise RuntimeError("Provided trial does not have a compatible configuration. "
                           "A dimension named '%s' should be present.\n %s" %
                           (name, trial))

    return False


class DimensionAddition(BaseAdapter):
    """Adapter which adds a new dimension to parent's trials

    This adaptation is based on the assumption that the trials of the parent experiment
    are equivalent if we add the new dimension for a given default value.

    On forward, the adapter add a dimension with provided default value to each trials.
    On backward, the adapter filters trials and only keep with with the default value.

    Attributes
    ----------
    param: instances of `orion.core.worker.trial.Trial.Param`
        A parameter object which defines the name and default value of the name dimension.

    """

    def __init__(self, param):
        """Initialize and instantiate the param if necessary

        Parameters
        ----------
        param: instance of `orion.core.worker.trial.Trial.Param` or `dict`
            A parameter object which defines the name and default value of the name dimension.
            It can be either a dictionary definition or an instance of Param. If the former, then
            a parameter is instantiated from the dictionary.

        """
        if isinstance(param, dict):
            param = Trial.Param(**param)
        elif not isinstance(param, Trial.Param):
            raise TypeError("Invalid param argument type ('%s'). "
                            "Param argument must be a Param object or a dictionnary "
                            "as defined by Trial.Param.to_dict()." %
                            str(type(param)))

        self.param = param

    def forward(self, trials):
        """Add a dimension with provided default value to each trials of the parent

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        adapted_trials = []

        for trial in trials:
            if apply_if_valid(self.param.name, trial, raise_if_not=False):
                raise RuntimeError("Provided trial does not have a compatible configuration. "
                                   "A dimension named '%s' was already present.\n %s" %
                                   (self.param.name, trial))

            adapted_trial = copy.deepcopy(trial)
            # pylint: disable=protected-access
            adapted_trial._params.append(copy.deepcopy(self.param))
            adapted_trials.append(adapted_trial)

        return adapted_trials

    def backward(self, trials):
        """Filter out trials which have values different than the default one.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        adapted_trials = []

        def remove_dimension(trial, param):
            """Remove the param and keep the trial if param has default value"""
            if param == self.param:
                adapted_trial = copy.deepcopy(trial)
                # pylint: disable=protected-access
                del adapted_trial._params[adapted_trial._params.index(self.param)]
                adapted_trials.append(adapted_trial)
                return True

            return False

        for trial in trials:
            apply_if_valid(self.param.name, trial, remove_dimension, raise_if_not=True)

        return adapted_trials

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower(),
            param=self.param.to_dict())
        return ret


class DimensionDeletion(BaseAdapter):
    """Adapter which remove a dimension to parent's trials

    .. note::

        This adapter is the opposite of `orion.core.evc.adapters.DimensionAddition`.

    This adaptation is based on the assumption that the trials of the children experiment
    are equivalent if we remove the new dimension for a given default value.

    On forward, the adapter filters trials and only keep with with the default value.
    On backward, the adapter add a dimension with provided default value to each trials.

    Attributes
    ----------
    dimension_addition_adapter: instance of `orion.core.evc.adapters.DimensionAddition`
        An adapter to add a new dimension, it is used by DimensionDeletion inversely to remove
        a dimension.

    """

    def __init__(self, param):
        """Initialize and instantiate the param if necessary

        Parameters
        ----------
        param: instance of `orion.core.worker.trial.Trial.Param` or `dict`
            A parameter object which defines the name and default value of the name dimension.
            It can be either a dictionary definition or an instance of Param. If the former, then
            a parameter is instantiated from the dictionary.

        """
        self.dimension_addition_adapter = DimensionAddition(param)

    @property
    def param(self):
        """Parameter containing name and default value"""
        return self.dimension_addition_adapter.param

    def forward(self, trials):
        """Filter out trials which have values different than the default one.

        .. seealso::

            :meth:`orion.core.evc.DimensionAddition.backward`
            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        return self.dimension_addition_adapter.backward(trials)

    def backward(self, trials):
        """Add a dimension with provided default value to each trials of the child.

        .. seealso::

            :meth:`orion.core.evc.DimensionAddition.forward`
            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        return self.dimension_addition_adapter.forward(trials)

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = self.dimension_addition_adapter.to_dict()
        ret['of_type'] = 'dimensiondeletion'
        return ret


class DimensionPriorChange(BaseAdapter):
    """Adapter which filters parent's trials based on a new prior

    On forward, the adapter filters out trials based on the child's prior.
    On backward, the adapter filters out trials based on the parent's prior.

    Attributes
    ----------
    name: `str`
        Name of the dimension.
    old_prior: `str`
        string definition as parsable by `orion.core.io.space_builder`.
    new_prior: `str`
        string definition as parsable by `orion.core.io.space_builder`.
    old_dimension: instance of `orion.algo.space.Dimension`
        The dimension of the parent experiment
    new_dimension: instance of `orion.algo.space.Dimension`
        The dimension of the child experiment

    """

    def __init__(self, name, old_prior, new_prior):
        """Initialize and instantiate dimensions

        Parameters
        ----------
        name: `str`
            Name of the dimension.
        old_prior: `str`
            string definition as parsable by `orion.core.io.space_builder`.
        new_prior: `str`
            string definition as parsable by `orion.core.io.space_builder`.

        """
        self.name = name
        self.old_prior = old_prior
        self.new_prior = new_prior
        self.old_dimension = DimensionBuilder().build('old', old_prior)
        self.new_dimension = DimensionBuilder().build('new', new_prior)

        if self.old_dimension.shape != self.new_dimension.shape:
            raise NotImplementedError("Oríon does not support yet adaptations on prior "
                                      "shape changes.")

    def forward(self, trials):
        """Filter out trials which have out of bound values based on the child's prior.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        # pylint: disable=unused-argument
        def is_in_bound(trial, param):
            """Test if param's value is in the new prior's bounds"""
            return param.value in self.new_dimension

        return [trial for trial in trials if apply_if_valid(self.name, trial, callback=is_in_bound)]

    def backward(self, trials):
        """Filter out trials which have out of bound values based on the parent's prior.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        return DimensionPriorChange(self.name, self.new_prior, self.old_prior).forward(trials)

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower(),
            name=self.name,
            old_prior=self.old_prior,
            new_prior=self.new_prior)
        return ret


class DimensionRenaming(BaseAdapter):
    """Adapter which change the name of a dimension in parent's trials

    On forward, the adapter change dimensions name from A to B.
    On backward, the adapter change dimensions name from B to A.

    Attributes
    ----------
    old_name: `str`
        Name of the parent's dimension.
    new_name: `str`
        Name of the child's dimension.

    """

    def __init__(self, old_name, new_name):
        """Initialize

        Parameters
        ----------
        old_name: `str`
            Name of the parent's dimension.
        new_name: `str`
            Name of the child's dimension.

        """
        if any(not isinstance(name, str) for name in [old_name, new_name]):
            wrong_object_type = [type(name) for name in [old_name, new_name]
                                 if not isinstance(name, str)][0]
            raise TypeError("Invalid name type '%s'. Names must be strings." %
                            str(wrong_object_type))
        self.old_name = old_name
        self.new_name = new_name

    def forward(self, trials):
        """Change name of dimension `old_name` to `new_name`.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        # pylint: disable=unused-argument
        def rename(trial, param):
            """Rename param to given new name"""
            param.name = self.new_name
            return True

        adapted_trials = copy.deepcopy(trials)

        for trial in adapted_trials:
            apply_if_valid(self.old_name, trial, callback=rename, raise_if_not=True)

        return adapted_trials

    def backward(self, trials):
        """Change name of dimension `new_name` to `old_name`.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        return DimensionRenaming(self.new_name, self.old_name).forward(trials)

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower(),
            old_name=self.old_name,
            new_name=self.new_name)
        return ret


class AlgorithmChange(BaseAdapter):
    """Adapter for changes in algorithm definition

    .. note::

        Current implementation does nothing

    """

    def forward(self, trials):
        """Pass all trials from parent experiment to child experiment

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        return trials

    def backward(self, trials):
        """Pass all trials from child experiment to parent experiment

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        return trials

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower())
        return ret


class CodeChange(BaseAdapter):
    """Adapter which filters parent's trials based on the type of code change

    This adapter let pass all trials if the code change didn't break the compatibility with parent's
    experiment. If the effect of the change is UNSURE, the trials may only pass from parent to child
    but not from child to parent. This is to ensure parent experiment does not get corrupted with
    possibly incompatible results.

    On forward, the adapter filters out parent's trials if type of code change is BREAK.
    On backward, the adapter filters out child's trials if type of code change is UNSURE or BREAK.

    Attributes
    ----------
    change_type: `str`
        Type of change of the code. Can be one of `CodeChange.NOEFFET`, `CodeChange.BREAK` or
        `CodeChange.UNSURE`.

    """

    NOEFFECT = "noeffect"
    BREAK = "break"
    UNSURE = "unsure"
    types = [NOEFFECT, BREAK, UNSURE]

    def __init__(self, change_type):
        """Initialize and check change type's validity

        Parameters
        ----------
        change_type: `str`
            Type of change of the code. Can be one of `CodeChange.NOEFFET`, `CodeChange.BREAK` or
            `CodeChange.UNSURE`.

        """
        self.validate(change_type)
        self.change_type = change_type

    @classmethod
    def validate(cls, change_type):
        """Validate change type and raise ValueError if invalid"""
        if change_type not in cls.types:
            raise ValueError("Invalid code change type '%s'. Should be one of %s" %
                             (change_type, str(cls.types)))

    def forward(self, trials):
        """Filter out parent's trials if type of code change is BREAK.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        if self.change_type == self.BREAK:
            return []

        return trials

    def backward(self, trials):
        """Filter out child's trials if type of code change is UNSURE or BREAK.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        if self.change_type in [self.BREAK, self.UNSURE]:
            return []

        return trials

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower(),
            change_type=self.change_type)
        return ret


class CommandLineChange(BaseAdapter):
    """Adapter which filters parent's trials based on the type of command line change

    This adapter let pass all trials if the command line change didn't break the compatibility with
    parent's experiment. If the effect of the change is UNSURE, the trials may only pass from parent
    to child but not from child to parent. This is to ensure parent experiment does not get
    corrupted with possibly incompatible results.

    On forward, the adapter filters out parent's trials if type of change is BREAK.
    On backward, the adapter filters out child's trials if type of change is UNSURE or BREAK.

    Attributes
    ----------
    change_type: `str`
        Type of change of the command line. Can be one of `CommandLineChange.NOEFFET`,
        `CommandLineChange.BREAK` or `CommandLineChange.UNSURE`.

    """

    NOEFFECT = "noeffect"
    BREAK = "break"
    UNSURE = "unsure"
    types = [NOEFFECT, BREAK, UNSURE]

    def __init__(self, change_type):
        """Initialize and check change type's validity

        Parameters
        ----------
        change_type: `str`
            Type of change of the command line. Can be one of `Change.NOEFFET`,
            `CommandLineChange.BREAK` or `CommandLineChange.UNSURE`.

        """
        self.validate(change_type)
        self.change_type = change_type

    @classmethod
    def validate(cls, change_type):
        """Validate change type and raise ValueError if invalid"""
        if change_type not in cls.types:
            raise ValueError("Invalid cli change type '%s'. Should be one of %s" %
                             (change_type, str(cls.types)))

    def forward(self, trials):
        """Filter out parent's trials if type of cli change is BREAK.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        if self.change_type == self.BREAK:
            return []

        return trials

    def backward(self, trials):
        """Filter out child's trials if type of cli change is UNSURE or BREAK.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        if self.change_type in [self.BREAK, self.UNSURE]:
            return []

        return trials

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower(),
            change_type=self.change_type)
        return ret


class ScriptConfigChange(BaseAdapter):
    """Adapter which filters parent's trials based on the type of user's script's config change

    This adapter let pass all trials if the change in the user's script's configuration file
    didn't break the compatibility with parent's experiment. If the effect of the change is UNSURE,
    the trials may only pass from parent to child but not from child to parent. This is to ensure
    parent experiment does not get corrupted with possibly incompatible results.

    On forward, the adapter filters out parent's trials if type of change is BREAK.
    On backward, the adapter filters out child's trials if type of change is UNSURE or BREAK.

    Attributes
    ----------
    change_type: `str`
        Type of change of the command line. Can be one of `ScriptConfigChange.NOEFFET`,
        `ScriptConfigChange.BREAK` or `ScriptConfigChange.UNSURE`.

    """

    NOEFFECT = "noeffect"
    BREAK = "break"
    UNSURE = "unsure"
    types = [NOEFFECT, BREAK, UNSURE]

    def __init__(self, change_type):
        """Initialize and check change type's validity

        Parameters
        ----------
        change_type: `str`
            Type of change of the script's config. Can be one of `ScriptConfigChange.NOEFFET`,
            `ScriptConfigChange.BREAK` or `ScriptConfigChange.UNSURE`.

        """
        self.validate(change_type)
        self.change_type = change_type

    @classmethod
    def validate(cls, change_type):
        """Validate change type and raise ValueError if invalid"""
        if change_type not in cls.types:
            raise ValueError("Invalid script's config change type '%s'. Should be one of %s" %
                             (change_type, str(cls.types)))

    def forward(self, trials):
        """Filter out parent's trials if type of script's config change is BREAK.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.forward`
        """
        if self.change_type == self.BREAK:
            return []

        return trials

    def backward(self, trials):
        """Filter out child's trials if type of script's config change is UNSURE or BREAK.

        .. seealso::

            :meth:`orion.core.evc.BaseAdapter.backward`
        """
        if self.change_type in [self.BREAK, self.UNSURE]:
            return []

        return trials

    def to_dict(self):
        """Provide the configuration of the adapter as a dictionary

        .. seealso::

            :meth:`orion.core.evc.adapters.BaseAdapter.to_dict`
        """
        ret = dict(
            of_type=self.__class__.__name__.lower(),
            change_type=self.change_type)
        return ret


# pylint: disable=too-few-public-methods,abstract-method
class Adapter(BaseAdapter, metaclass=Factory):
    """Class used to inject dependency on an adapter implementation.

    .. seealso:: `orion.core.utils.Factory` metaclass and `BaseAlgorithm` interface.
    """

    @classmethod
    def build(cls, adapter_dicts):
        """Builder method for a list of adapters.

        Parameters
        ----------
        adapter_dicts: list of `dict`
            List of adapter representation in dictionary form as expected to be saved in a database.

        Returns
        -------
        `orion.core.evc.adapters.CompositeAdapter`
            An adapter which may contain many adapters

        """
        adapters = []
        for adapter_dict in adapter_dicts:
            if isinstance(adapter_dict, (list, tuple)):
                adapter = Adapter.build(adapter_dict)
            else:
                adapter = cls(**adapter_dict)
            adapters.append(adapter)

        return CompositeAdapter(*adapters)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.storage`."""

import copy
import datetime
import logging
import time

import pytest

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.io.database.pickleddb import PickledDB
from orion.core.utils.singleton import SingletonAlreadyInstantiatedError, \
    SingletonNotInstantiatedError, update_singletons
from orion.core.worker.trial import Trial
from orion.storage.base import FailedUpdate, get_storage, MissingArguments, setup_storage, Storage
from orion.storage.legacy import Legacy
from orion.storage.track import HAS_TRACK, REASON
from orion.testing import OrionState

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)

storage_backends = [
    None  # defaults to legacy with PickleDB
]

if not HAS_TRACK:
    log.warning('Track is not tested because: %s!', REASON)
else:
    storage_backends.append({
        'type': 'track',
        'uri': 'file://${file}?objective=loss'
    })

base_experiment = {
    'name': 'default_name',
    'version': 0,
    'metadata': {
        'user': 'default_user',
        'user_script': 'abc',
        'priors': {'x': 'uniform(0, 10)'},
        'datetime': '2017-11-23T02:00:00'
    }
}


base_trial = {
    'experiment': 'default_name',
    'status': 'new',  # new, reserved, suspended, completed, broken
    'worker': None,
    'submit_time': '2017-11-23T02:00:00',
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [
        {'name': 'loss',
         'type': 'objective',  # objective, constraint
         'value': 2}
    ],
    'params': [
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'},
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'lstm_with_attention'}
    ]
}


def _generate(obj, *args, value):
    if obj is None:
        return None

    obj = copy.deepcopy(obj)
    data = obj

    for arg in args[:-1]:
        data = data[arg]

    data[args[-1]] = value
    return obj


def make_lost_trial(delay=2):
    """Make a lost trial"""
    obj = copy.deepcopy(base_trial)
    obj['status'] = 'reserved'
    obj['heartbeat'] = (datetime.datetime.utcnow() -
                        datetime.timedelta(seconds=orion.core.config.worker.heartbeat * delay))
    obj['params'].append({
        'name': '/index',
        'type': 'categorical',
        'value': f'lost_trial_{delay}'
    })
    return obj


all_status = ['completed', 'broken', 'reserved', 'interrupted', 'suspended', 'new']


def generate_trials(status=None, heartbeat=None):
    """Generate Trials with different configurations"""
    if status is None:
        status = all_status

    new_trials = [_generate(base_trial, 'status', value=s) for s in status]

    if heartbeat:
        for trial in new_trials:
            trial['heartbeat'] = heartbeat

    # make each trial unique
    for i, trial in enumerate(new_trials):
        trial['params'].append({
            'name': '/index',
            'type': 'categorical',
            'value': i
        })

    return new_trials


def generate_experiments():
    """Generate a set of experiments"""
    users = ['a', 'b', 'c']
    exps = [_generate(base_experiment, 'metadata', 'user', value=u) for u in users]
    return [_generate(exp, 'name', value=str(i)) for i, exp in enumerate(exps)]


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_setup_storage_default():
    """Test that storage is setup using default config"""
    update_singletons()
    setup_storage()
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)


def test_setup_storage_bad():
    """Test how setup fails when configuring with non-existant backends"""
    update_singletons()
    with pytest.raises(NotImplementedError) as exc:
        setup_storage({'type': 'idontexist'})

    assert exc.match('idontexist')


def test_setup_storage_custom():
    """Test setup with local configuration"""
    update_singletons()
    setup_storage({'type': 'legacy', 'database': {'type': 'pickleddb', 'host': 'test.pkl'}})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert storage._db.host == 'test.pkl'


def test_setup_storage_custom_type_missing():
    """Test setup with local configuration with type missing"""
    update_singletons()
    setup_storage({'database': {'type': 'pickleddb', 'host': 'test.pkl'}})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert storage._db.host == 'test.pkl'


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_setup_storage_custom_legacy_emtpy():
    """Test setup with local configuration with legacy but no config"""
    update_singletons()
    setup_storage({'type': 'legacy'})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert storage._db.host == orion.core.config.storage.database.host


def test_setup_storage_bad_override():
    """Test setup with different type than existing singleton"""
    update_singletons()
    setup_storage({'type': 'legacy', 'database': {'type': 'pickleddb', 'host': 'test.pkl'}})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    with pytest.raises(SingletonAlreadyInstantiatedError) as exc:
        setup_storage({'type': 'track'})

    assert exc.match('A singleton instance of \(type: Storage\)')


@pytest.mark.xfail(reason='Fix this when introducing #135 in v0.2.0')
def test_setup_storage_bad_config_override():
    """Test setup with different config than existing singleton"""
    update_singletons()
    setup_storage({'database': {'type': 'pickleddb', 'host': 'test.pkl'}})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    with pytest.raises(SingletonAlreadyInstantiatedError):
        setup_storage({'database': {'type': 'mongodb'}})


def test_setup_storage_stateless():
    """Test that passed configuration dictionary is not modified by the fonction"""
    update_singletons()
    config = {'database': {'type': 'pickleddb', 'host': 'test.pkl'}}
    passed_config = copy.deepcopy(config)
    setup_storage(passed_config)
    assert config == passed_config


def test_get_storage_uninitiated():
    """Test that get storage fails if no storage singleton exist"""
    update_singletons()
    with pytest.raises(SingletonNotInstantiatedError) as exc:
        get_storage()

    assert exc.match('No singleton instance of \(type: Storage\) was created')


def test_get_storage():
    """Test that get storage gets the singleton"""
    update_singletons()
    setup_storage({'database': {'type': 'pickleddb', 'host': 'test.pkl'}})
    storage = get_storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert get_storage() == storage


@pytest.mark.parametrize('storage', storage_backends)
class TestStorage:
    """Test all storage backend"""

    def test_create_experiment(self, storage):
        """Test create experiment"""
        with OrionState(experiments=[], storage=storage) as cfg:
            storage = cfg.storage()

            storage.create_experiment(base_experiment)

            experiments = storage.fetch_experiments({})
            assert len(experiments) == 1, 'Only one experiment in the database'

            experiment = experiments[0]
            assert base_experiment == experiment, 'Local experiment and DB should match'

            # Insert it again
            with pytest.raises(DuplicateKeyError):
                storage.create_experiment(base_experiment)

    def test_fetch_experiments(self, storage, name='0', user='a'):
        """Test fetch experiments"""
        with OrionState(experiments=generate_experiments(), storage=storage) as cfg:
            storage = cfg.storage()

            experiments = storage.fetch_experiments({})
            assert len(experiments) == len(cfg.experiments)

            experiments = storage.fetch_experiments({'name': name, 'metadata.user': user})
            assert len(experiments) == 1

            experiment = experiments[0]
            assert experiment['name'] == name, 'name should match query'
            assert experiment['metadata']['user'] == user, 'user name should match query'

            experiments = storage.fetch_experiments({'name': '-1', 'metadata.user': user})
            assert len(experiments) == 0

    def test_update_experiment(self, monkeypatch, storage, name='0', user='a'):
        """Test fetch experiments"""
        with OrionState(experiments=generate_experiments(), storage=storage) as cfg:
            storage = cfg.storage()

            class _Dummy:
                pass

            experiment = cfg.experiments[0]
            mocked_experiment = _Dummy()
            mocked_experiment.id = experiment['_id']

            storage.update_experiment(mocked_experiment, test=True)
            experiments = storage.fetch_experiments({'_id': experiment['_id']})
            assert len(experiments) == 1

            fetched_experiment = experiments[0]
            assert fetched_experiment['test']

            assert 'test' not in storage.fetch_experiments({'_id': cfg.experiments[1]['_id']})[0]

            storage.update_experiment(uid=experiment['_id'], test2=True)
            assert storage.fetch_experiments({'_id': experiment['_id']})[0]['test2']
            assert 'test2' not in storage.fetch_experiments({'_id': cfg.experiments[1]['_id']})[0]

            with pytest.raises(MissingArguments):
                storage.update_experiment()

            with pytest.raises(AssertionError):
                storage.update_experiment(experiment=mocked_experiment, uid='123')

    def test_delete_experiment(self, storage):
        """Test delete one experiment"""
        if storage and storage['type'] == 'track':
            pytest.xfail("Track does not support deletion yet.")

        with OrionState(experiments=generate_experiments(), storage=storage) as cfg:
            storage = cfg.storage()

            n_experiments = len(storage.fetch_experiments({}))
            storage.delete_experiment(uid=cfg.experiments[0]['_id'])
            experiments = storage.fetch_experiments({})
            assert len(experiments) == n_experiments - 1
            assert cfg.experiments[0]['_id'] not in [exp['_id'] for exp in experiments]

    def test_register_trial(self, storage):
        """Test register trial"""
        with OrionState(experiments=[base_experiment], storage=storage) as cfg:
            storage = cfg.storage()
            trial1 = storage.register_trial(Trial(**base_trial))
            trial2 = storage.get_trial(trial1)

            assert trial1.to_dict() == trial2.to_dict(), 'Trials should match after insert'

    def test_register_duplicate_trial(self, storage):
        """Test register trial"""
        with OrionState(
                experiments=[base_experiment], trials=[base_trial], storage=storage) as cfg:
            storage = cfg.storage()

            with pytest.raises(DuplicateKeyError):
                storage.register_trial(Trial(**base_trial))

    def test_register_lie(self, storage):
        """Test register lie"""
        with OrionState(experiments=[base_experiment], storage=storage) as cfg:
            storage = cfg.storage()
            storage.register_lie(Trial(**base_trial))

    def test_register_lie_fail(self, storage):
        """Test register lie"""
        with OrionState(experiments=[base_experiment], lies=[base_trial], storage=storage) as cfg:
            storage = cfg.storage()

            with pytest.raises(DuplicateKeyError):
                storage.register_lie(Trial(**cfg.lies[0]))

    def test_update_trials(self, storage):
        """Test update many trials"""
        with OrionState(
                experiments=[base_experiment],
                trials=generate_trials(status=['completed', 'reserved', 'reserved']),
                storage=storage) as cfg:
            storage = cfg.storage()

            class _Dummy:
                pass

            experiment = cfg.get_experiment('default_name', version=None)
            trials = storage.fetch_trials(experiment)
            assert sum(trial.status == 'reserved' for trial in trials) == 2
            count = storage.update_trials(
                experiment,
                where={'status': 'reserved'},
                status='interrupted')
            assert count == 2
            trials = storage.fetch_trials(experiment)
            assert sum(trial.status == 'interrupted' for trial in trials) == 2

    def test_update_trial(self, storage):
        """Test update one trial"""
        with OrionState(experiments=[base_experiment], trials=[base_trial], storage=storage) as cfg:
            storage = cfg.storage()

            trial = Trial(**cfg.trials[0])

            assert trial.status != 'interrupted'
            storage.update_trial(trial, status='interrupted')
            assert storage.get_trial(trial).status == 'interrupted'

    def test_reserve_trial_success(self, storage):
        """Test reserve trial"""
        with OrionState(
                experiments=[base_experiment], trials=[base_trial], storage=storage) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment('default_name', version=None)

            trial = storage.reserve_trial(experiment)

            assert trial is not None
            assert trial.status == 'reserved'

    def test_reserve_trial_fail(self, storage):
        """Test reserve trial"""
        with OrionState(
                experiments=[base_experiment],
                trials=generate_trials(status=['completed', 'reserved']),
                storage=storage) as cfg:

            storage = cfg.storage()
            experiment = cfg.get_experiment('default_name', version=None)

            trial = storage.reserve_trial(experiment)
            assert trial is None

    def test_fetch_trials(self, storage):
        """Test fetch experiment trials"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment('default_name', version=None)

            trials1 = storage.fetch_trials(experiment=experiment)
            trials2 = storage.fetch_trials(uid=experiment.id)

            with pytest.raises(MissingArguments):
                storage.fetch_trials()

            with pytest.raises(AssertionError):
                storage.fetch_trials(experiment=experiment, uid='123')

            assert len(trials1) == len(cfg.trials), 'trial count should match'
            assert len(trials2) == len(cfg.trials), 'trial count should match'

    def test_delete_all_trials(self, storage):
        """Test delete all trials of an experiment"""
        if storage and storage['type'] == 'track':
            pytest.xfail("Track does not support deletion yet.")

        trials = generate_trials()
        trial_from_other_exp = copy.deepcopy(trials[0])
        trial_from_other_exp['experiment'] = 'other'
        trials.append(trial_from_other_exp)
        with OrionState(
                experiments=[base_experiment], trials=trials, storage=storage) as cfg:
            storage = cfg.storage()

            # Make sure we have sufficient trials to test deletion
            trials = storage.fetch_trials(uid='default_name')
            assert len(trials) > 2

            count = storage.delete_trials(uid='default_name')
            assert count == len(trials)
            assert storage.fetch_trials(uid='default_name') == []

            # Make sure trials from other experiments were not deleted
            assert len(storage.fetch_trials(uid='other')) == 1

    def test_delete_trials_with_query(self, storage):
        """Test delete experiment trials matching a query"""
        if storage and storage['type'] == 'track':
            pytest.xfail("Track does not support deletion yet.")

        trials = generate_trials()
        trial_from_other_exp = copy.deepcopy(trials[0])
        trial_from_other_exp['experiment'] = 'other'
        trials.append(trial_from_other_exp)
        with OrionState(
                experiments=[base_experiment], trials=trials, storage=storage) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment('default_name')

            # Make sure we have sufficient trials to test deletion
            status = trials[0]['status']
            trials = storage.fetch_trials(experiment)
            trials_with_status = storage.fetch_trials_by_status(experiment, status)
            assert len(trials_with_status) > 0
            assert len(trials) > len(trials_with_status)

            # Test deletion
            count = storage.delete_trials(uid='default_name', where={'status': status})
            assert count == len(trials_with_status)
            assert storage.fetch_trials_by_status(experiment, status) == []
            assert len(storage.fetch_trials(experiment)) == len(trials) - len(trials_with_status)

            # Make sure trials from other experiments were not deleted
            assert len(storage.fetch_trials(uid='other')) == 1

    def test_get_trial(self, storage):
        """Test get trial"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            storage = cfg.storage()

            trial_dict = cfg.trials[0]

            trial1 = storage.get_trial(trial=Trial(**trial_dict))
            trial2 = storage.get_trial(uid=trial1.id)

            with pytest.raises(MissingArguments):
                storage.get_trial()

            with pytest.raises(AssertionError):
                storage.get_trial(trial=trial1, uid='123')

            assert trial1.to_dict() == trial_dict
            assert trial2.to_dict() == trial_dict

    def test_fetch_lost_trials(self, storage):
        """Test update heartbeat"""
        lost_trials = [make_lost_trial(2),  # Is not lost for long enough to be catched
                       make_lost_trial(10)]  # Is lost for long enough to be catched
        # Force recent heartbeat to avoid mixing up with lost trials.
        trials = lost_trials + generate_trials(heartbeat=datetime.datetime.utcnow())
        with OrionState(experiments=[base_experiment], trials=trials, storage=storage) as cfg:
            storage = cfg.storage()

            experiment = cfg.get_experiment('default_name', version=None)
            trials = storage.fetch_lost_trials(experiment)

            assert len(trials) == 1

    def test_change_status_success(self, storage):
        """Change the status of a Trial"""
        def check_status_change(new_status):
            with OrionState(
                    experiments=[base_experiment],
                    trials=generate_trials(), storage=storage) as cfg:
                trial = get_storage().get_trial(cfg.get_trial(0))
                assert trial is not None, 'was not able to retrieve trial for test'

                get_storage().set_trial_status(trial, status=new_status)
                assert trial.status == new_status, \
                    'Trial status should have been updated locally'

                trial = get_storage().get_trial(trial)
                assert trial.status == new_status, \
                    'Trial status should have been updated in the storage'

        check_status_change('completed')
        check_status_change('broken')
        check_status_change('reserved')
        check_status_change('interrupted')
        check_status_change('suspended')
        check_status_change('new')

    def test_change_status_invalid(self, storage):
        """Attempt to change the status of a Trial with an invalid one"""
        with OrionState(
                experiments=[base_experiment],
                trials=generate_trials(), storage=storage) as cfg:
            trial = get_storage().get_trial(cfg.get_trial(0))
            assert trial is not None, 'Was not able to retrieve trial for test'

            with pytest.raises(ValueError) as exc:
                get_storage().set_trial_status(trial, status='moo')

            assert exc.match('Given status `moo` not one of')

    def test_change_status_failed_update(self, storage):
        """Change the status of a Trial"""
        def check_status_change(new_status):
            with OrionState(
                    experiments=[base_experiment],
                    trials=generate_trials(), storage=storage) as cfg:
                trial = get_storage().get_trial(cfg.get_trial(0))
                assert trial is not None, 'Was not able to retrieve trial for test'
                assert trial.status != new_status

                if trial.status == new_status:
                    return

                with pytest.raises(FailedUpdate):
                    trial.status = new_status
                    get_storage().set_trial_status(trial, status=new_status)

        check_status_change('completed')
        check_status_change('broken')
        check_status_change('reserved')
        check_status_change('interrupted')
        check_status_change('suspended')

    def test_fetch_pending_trials(self, storage):
        """Test fetch pending trials"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            storage = cfg.storage()

            experiment = cfg.get_experiment('default_name', version=None)
            trials = storage.fetch_pending_trials(experiment)

            count = 0
            for trial in cfg.trials:
                if trial['status'] in {'new', 'suspended', 'interrupted'}:
                    count += 1

            assert len(trials) == count
            for trial in trials:
                assert trial.status in {'new', 'suspended', 'interrupted'}

    def test_fetch_noncompleted_trials(self, storage):
        """Test fetch non completed trials"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            storage = cfg.storage()

            experiment = cfg.get_experiment('default_name', version=None)
            trials = storage.fetch_noncompleted_trials(experiment)

            count = 0
            for trial in cfg.trials:
                if trial['status'] != 'completed':
                    count += 1

            for trial in trials:
                assert trial.status != 'completed'

            assert len(trials) == count

    def test_fetch_trials_by_status(self, storage):
        """Test fetch completed trials"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            count = 0
            for trial in cfg.trials:
                if trial['status'] == 'completed':
                    count += 1

            storage = cfg.storage()
            experiment = cfg.get_experiment('default_name', version=None)
            trials = storage.fetch_trials_by_status(experiment, 'completed')

            assert len(trials) == count
            for trial in trials:
                assert trial.status == 'completed', trial

    def test_count_completed_trials(self, storage):
        """Test count completed trials"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            count = 0
            for trial in cfg.trials:
                if trial['status'] == 'completed':
                    count += 1

            storage = cfg.storage()

            experiment = cfg.get_experiment('default_name', version=None)
            trials = storage.count_completed_trials(experiment)
            assert trials == count

    def test_count_broken_trials(self, storage):
        """Test count broken trials"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            count = 0
            for trial in cfg.trials:
                if trial['status'] == 'broken':
                    count += 1

            storage = cfg.storage()

            experiment = cfg.get_experiment('default_name', version=None)

            trials = storage.count_broken_trials(experiment)

            assert trials == count

    def test_update_heartbeat(self, storage):
        """Test update heartbeat"""
        with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage) as cfg:
            storage_name = storage
            storage = cfg.storage()

            exp = cfg.get_experiment('default_name')
            trial1 = storage.fetch_trials_by_status(exp, status='reserved')[0]
            trial1b = copy.deepcopy(trial1)

            storage.update_heartbeat(trial1)

            trial2 = storage.get_trial(trial1)

            # this check that heartbeat is the correct type and that it was updated prior to now
            assert trial1b.heartbeat is None
            assert trial1.heartbeat is None
            assert trial2.heartbeat is not None

            # Sleep a bit, because fast CPUs make this test fail
            time.sleep(0.1)
            assert trial2.heartbeat < datetime.datetime.utcnow()

            if storage_name is None:
                trial3 = storage.fetch_trials_by_status(exp, status='completed')[0]
                storage.update_heartbeat(trial3)

                assert trial3.heartbeat is None, \
                    'Legacy does not update trials with a status different from reserved'



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.diff` -- Utilitary functions for diffs
=============================================================

.. module:: experiment
   :platform: Unix
   :synopsis: Utilitary functions to provided colored diffs of (multi-line) strings

"""

from difflib import Differ


def red(string):
    """Return the string wrapped in code to color it in red in user's terminal"""
    return color(string, "31m")


def green(string):
    """Return the string wrapped in code to color it in green in user's terminal"""
    return color(string, "32m")


def color(string, color_code):
    """Return the string wrapped in code to color it in given `color_code` in user's terminal"""
    return "\x1b[0;" + color_code + string + "\x1b[0m"


def colored_diff(string_a, string_b):
    """Return a multi-line colored diff of given strings `a` and `b`"""
    string_a = string_a.splitlines(keepends=True)
    string_b = string_b.splitlines(keepends=True)
    result = list(Differ().compare(string_a, string_b))

    colored_result = []
    for i, line in enumerate(result):
        line = line.strip('\n')
        if line[0] == "-" or (line[0] == "?" and result[i - 1][0] == "-"):
            line = red(line)
        elif line[0] == "+" or (line[0] == "?" and result[i - 1][0] == "+"):
            line = green(line)

        colored_result.append(line)

    return "\n".join(colored_result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures and utils for unittests and functional tests."""
import os
import tempfile

import numpy
from pymongo import MongoClient
import pytest
import yaml

from orion.algo.base import (BaseAlgorithm, OptimizationAlgorithm)
import orion.core
from orion.core.io import resolve_config
from orion.core.io.database import Database
from orion.core.io.database.mongodb import MongoDB
from orion.core.io.database.pickleddb import PickledDB
import orion.core.utils.backward as backward
from orion.core.utils.singleton import update_singletons
from orion.core.worker.trial import Trial
from orion.storage.base import Storage
from orion.storage.legacy import Legacy
from orion.testing import OrionState


@pytest.fixture(scope="session", autouse=True)
def shield_from_user_config(request):
    """Do not read user's yaml global config."""
    _pop_out_yaml_from_config(orion.core.config)


def _pop_out_yaml_from_config(config):
    """Remove any configuration fetch from yaml file"""
    for key in config._config.keys():
        config._config[key].pop('yaml', None)

    for key in config._subconfigs.keys():
        _pop_out_yaml_from_config(config._subconfigs[key])


class DumbAlgo(BaseAlgorithm):
    """Stab class for `BaseAlgorithm`."""

    def __init__(self, space, value=5,
                 scoring=0, judgement=None,
                 suspend=False, done=False, seed=None, **nested_algo):
        """Configure returns, allow for variable variables."""
        self._times_called_suspend = 0
        self._times_called_is_done = 0
        self._num = 0
        self._index = 0
        self._points = []
        self._suggested = None
        self._results = []
        self._score_point = None
        self._judge_point = None
        self._measurements = None
        self.possible_values = [value]
        super(DumbAlgo, self).__init__(space, value=value,
                                       scoring=scoring, judgement=judgement,
                                       suspend=suspend,
                                       done=done,
                                       seed=seed,
                                       **nested_algo)

    def seed(self, seed):
        """Set the index to seed.

        Setting the seed as an index so that unit-tests can force the algorithm to suggest the same
        values as if seeded.
        """
        self._index = seed if seed is not None else 0

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        _state_dict = super(DumbAlgo, self).state_dict
        _state_dict.update({'index': self._index, 'suggested': self._suggested, 'num': self._num,
                            'done': self.done})
        return _state_dict

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        super(DumbAlgo, self).set_state(state_dict)
        self._index = state_dict['index']
        self._suggested = state_dict['suggested']
        self._num = state_dict['num']
        self.done = state_dict['done']

    def suggest(self, num=1):
        """Suggest based on `value`."""
        self._num += num

        rval = []
        while len(rval) < num:
            value = self.possible_values[min(self._index, len(self.possible_values) - 1)]
            self._index += 1
            rval.append(value)

        self._suggested = rval

        return rval

    def observe(self, points, results):
        """Log inputs."""
        super(DumbAlgo, self).observe(points, results)
        self._points += points
        self._results += results

    def score(self, point):
        """Log and return stab."""
        self._score_point = point
        return self.scoring

    def judge(self, point, measurements):
        """Log and return stab."""
        self._judge_point = point
        self._measurements = measurements
        return self.judgement

    @property
    def should_suspend(self):
        """Cound how many times it has been called and return `suspend`."""
        self._times_called_suspend += 1
        return self.suspend

    @property
    def is_done(self):
        """Cound how many times it has been called and return `done`."""
        self._times_called_is_done += 1
        return self.done


# Hack it into being discoverable
OptimizationAlgorithm.types.append(DumbAlgo)
OptimizationAlgorithm.typenames.append(DumbAlgo.__name__.lower())


@pytest.fixture()
def empty_config():
    """Return config purged from global definition"""
    orion.core.DEF_CONFIG_FILES_PATHS = []
    config = orion.core.build_config()
    orion.core.config = config
    resolve_config.config = config
    return config


@pytest.fixture()
def test_config(empty_config):
    """Return orion's config overwritten with local config file"""
    config_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               "orion_config.yaml")
    empty_config.load_yaml(config_file)

    return empty_config


@pytest.fixture(scope='session')
def dumbalgo():
    """Return stab algorithm class."""
    return DumbAlgo


@pytest.fixture()
def categorical_values():
    """Return a list of all the categorical points possible for `supernaedo2` and `supernaedo3`"""
    return [('rnn', 'rnn'), ('lstm_with_attention', 'rnn'), ('gru', 'rnn'),
            ('rnn', 'gru'), ('lstm_with_attention', 'gru'), ('gru', 'gru'),
            ('rnn', 'lstm'), ('lstm_with_attention', 'lstm'), ('gru', 'lstm')]


@pytest.fixture()
def exp_config_file():
    """Return configuration file used for stuff"""
    return os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        'unittests', 'core', 'experiment.yaml')


@pytest.fixture()
def exp_config():
    """Load an example database."""
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
              'unittests', 'core', 'experiment.yaml')) as f:
        exp_config = list(yaml.safe_load_all(f))

    for i, t_dict in enumerate(exp_config[1]):
        exp_config[1][i] = Trial(**t_dict).to_dict()

    for config in exp_config[0]:
        config["metadata"]["user_script"] = os.path.join(
            os.path.dirname(__file__), config["metadata"]["user_script"])
        backward.populate_space(config)
        config['version'] = 1

    return exp_config


@pytest.fixture(scope='session')
def database():
    """Return Mongo database object to test with example entries."""
    client = MongoClient(username='user', password='pass', authSource='orion_test')
    database = client.orion_test
    yield database
    client.close()


@pytest.fixture()
def mock_database():
    """
    Lightweight fixture for an empty, in-memory database using :class:`OrionState`.
    The database is automatically discarded after each test method.
    """
    storage = {'type': 'legacy', 'database': {'type': 'EphemeralDB'}}
    with OrionState(storage=storage) as state:
        yield state


@pytest.fixture()
def clean_db(database, exp_config):
    """Clean insert example experiment entries to collections."""
    database.experiments.drop()
    database.experiments.insert_many(exp_config[0])
    database.lying_trials.drop()
    database.trials.drop()
    database.trials.insert_many(exp_config[1])
    database.workers.drop()
    database.workers.insert_many(exp_config[2])
    database.resources.drop()
    database.resources.insert_many(exp_config[3])


@pytest.fixture()
def null_db_instances():
    """Nullify singleton instance so that we can assure independent instantiation tests."""
    Storage.instance = None
    Legacy.instance = None
    Database.instance = None
    MongoDB.instance = None
    PickledDB.instance = None


@pytest.fixture(scope='function')
def seed():
    """Return a fixed ``numpy.random.RandomState`` and global seed."""
    seed = 5
    rng = numpy.random.RandomState(seed)
    numpy.random.seed(seed)
    return rng


@pytest.fixture
def version_XYZ(monkeypatch):
    """Force orion version XYZ on output of resolve_config.fetch_metadata"""
    non_patched_fetch_metadata = resolve_config.fetch_metadata

    def fetch_metadata(user=None, user_args=None):
        metadata = non_patched_fetch_metadata(user, user_args)
        metadata['orion_version'] = 'XYZ'
        return metadata
    monkeypatch.setattr(resolve_config, "fetch_metadata", fetch_metadata)


@pytest.fixture()
def create_db_instance(null_db_instances, clean_db):
    """Create and save a singleton database instance."""
    try:
        database = {
            'type': 'MongoDB',
            'name': 'orion_test',
            'username': 'user',
            'password': 'pass'
        }
        db = Storage(of_type='legacy', database=database)
    except ValueError:
        db = Storage()

    db = db._db
    return db


@pytest.fixture()
def script_path():
    """Return a script path for mock"""
    return os.path.join(os.path.dirname(__file__), "functional/demo/black_box.py")


@pytest.fixture()
def mock_infer_versioning_metadata(monkeypatch):
    """Mock infer_versioning_metadata and create a VCS"""
    def fixed_dictionary(user_script):
        """Create VCS"""
        vcs = {}
        vcs['type'] = 'git'
        vcs['is_dirty'] = False
        vcs['HEAD_sha'] = "test"
        vcs['active_branch'] = None
        vcs['diff_sha'] = "diff"
        return vcs
    monkeypatch.setattr(resolve_config, "infer_versioning_metadata", fixed_dictionary)


@pytest.fixture(scope="function")
def setup_pickleddb_database():
    """Configure the database"""
    update_singletons()
    temporary_file = tempfile.NamedTemporaryFile()

    os.environ['ORION_DB_TYPE'] = "pickleddb"
    os.environ['ORION_DB_ADDRESS'] = temporary_file.name
    yield
    temporary_file.close()
    del os.environ['ORION_DB_TYPE']
    del os.environ['ORION_DB_ADDRESS']



# -*- coding: utf-8 -*-
# pylint:disable=protected-access
"""
:mod:`orion.core.evc.experiment` -- Experiment node for EVC
===========================================================

.. module:: experiment
   :platform: Unix
   :synopsis: Experiment nodes connecting experiments to the EVC tree

The experiments are connected to one another through the experiment nodes. The former can be created
standalone without an EVC tree. When connected to an `ExperimentNode`, the experiments gain access
to trials of other experiments by using method `ExperimentNode.fetch_trials`.

Helper functions are provided to fetch trials keeping the tree structure. Those can be helpful when
analyzing an EVC tree.

"""
import logging

from orion.core.evc.tree import TreeNode
from orion.storage.base import get_storage


log = logging.getLogger(__name__)


class ExperimentNode(TreeNode):
    """Experiment node to connect experiments to EVC tree.

    The node carries an experiment in attribute `item`. The node can be instantiated only using the
    name of the experiment. The experiment will be created lazily on access to `node.item`.

    Attributes
    ----------
    name: str
        Name of the experiment
    item: None or :class:`orion.core.worker.experiment.Experiment`
        None if the experiment is not initialized yet. When initializing lazily, it creates an
        `ExperimentView`.

    .. seealso::

        :py:class:`TreeNode` for tree-specific attributes and methods.

    """

    __slots__ = ('name', 'version', '_no_parent_lookup', '_no_children_lookup') + TreeNode.__slots__

    def __init__(self, name, version, experiment=None, parent=None, children=tuple()):
        """Initialize experiment node with item, experiment, parent and children

        .. seealso::
            :class:`orion.core.evc.tree.TreeNode` for information about the attributes
        """
        super(ExperimentNode, self).__init__(experiment, parent, children)
        self.name = name
        self.version = version

        self._no_parent_lookup = True
        self._no_children_lookup = True

    @property
    def item(self):
        """Get the experiment associated to the node

        Note that accessing `item` may trigger the lazy initialization of the experiment if it was
        not done already.
        """
        if self._item is None:
            # TODO: Find another way around the circular import
            import orion.core.io.experiment_builder as experiment_builder
            self._item = experiment_builder.build_view(name=self.name, version=self.version)
            self._item._experiment._node = self

        return self._item

    @property
    def parent(self):
        """Get parent of the experiment, None if no parent

        .. note::

            The instantiation of an EVC tree is lazy, which means accessing the parent of a node
            may trigger a call to database to build this parent live.

        """
        if self._parent is None and self._no_parent_lookup:
            self._no_parent_lookup = False
            query = {'_id': self.item.refers.get('parent_id')}
            selection = {'name': 1, 'version': 1}
            experiments = get_storage().fetch_experiments(query, selection)

            if experiments:
                parent = experiments[0]
                exp_node = ExperimentNode(name=parent['name'], version=parent.get('version', 1))
                self.set_parent(exp_node)
        return self._parent

    @property
    def children(self):
        """Get children of the experiment, empty list if no children

        .. note::

            The instantiation of an EVC tree is lazy, which means accessing the children of a node
            may trigger a call to database to build those children live.

        """
        if self._no_children_lookup:
            self._children = []
            self._no_children_lookup = False
            query = {'refers.parent_id': self.item.id}
            selection = {'name': 1, 'version': 1}
            experiments = get_storage().fetch_experiments(query, selection)
            for child in experiments:
                self.add_children(
                    ExperimentNode(name=child['name'],
                                   version=child.get('version', 1)))

        return self._children

    @property
    def adapter(self):
        """Get the adapter of the experiment with respect to its parent"""
        return self.item.refers["adapter"]

    @property
    def tree_name(self):
        """Return a formatted name of the Node for a tree pretty-print."""
        if self.item is not None:
            return self.name + "-v{}".format(self.item.version)

        return self.name

    def fetch_lost_trials(self):
        """See :meth:`orion.core.evc.experiment:Experiment._fetch_trials`"""
        return self._fetch_trials('fetch_lost_trials')

    def fetch_trials(self):
        """See :meth:`orion.core.evc.experiment:Experiment._fetch_trials`"""
        return self._fetch_trials('fetch_trials')

    def fetch_pending_trials(self):
        """See :meth:`orion.core.evc.experiment:Experiment._fetch_trials`"""
        return self._fetch_trials('fetch_pending_trials')

    def fetch_noncompleted_trials(self):
        """See :meth:`orion.core.evc.experiment:Experiment._fetch_trials`"""
        return self._fetch_trials('fetch_noncompleted_trials')

    def fetch_trials_by_status(self, status):
        """See :meth:`orion.core.evc.experiment:Experiment._fetch_trials`"""
        return self._fetch_trials('fetch_trials_by_status', status=status)

    def _fetch_trials(self, fun_name, *args, **kwargs):
        """Fetch trials recursively in the EVC tree using the fetch function `fun_name`.

        Parameters
        ----------
        fun_name: callable
            Function name to call to fetch trials. The function must be an attribute of
            :class:`orion.core.worker.experiment:Experiment`

        *args:
            Positional arguments to pass to `fun_name`.

        **kwargs
            Keyword arguments to pass to `fun_name.

        """
        def retrieve_trials(node, parent_or_children):
            """Retrieve the trials of a node/experiment."""
            fun = getattr(node.item, fun_name)
            # with_evc_tree needs to be False here or we will have an infinite loop
            trials = fun(*args, with_evc_tree=False, **kwargs)
            return dict(trials=trials, experiment=node.item), parent_or_children

        # get the trials of the parents
        parent_trials = None
        if self.parent is not None:
            parent_trials = self.parent.map(retrieve_trials, self.parent.parent)

        # get the trials of the children
        children_trials = self.map(retrieve_trials, self.children)
        children_trials.set_parent(parent_trials)

        adapt_trials(children_trials)
        return sum([node.item['trials'] for node in children_trials.root], [])


def _adapt_parent_trials(node, parent_trials_node):
    """Adapt trials from the parent recursively

    .. note::

        To call with node.map(fct, node.parent) to connect with parents

    """
    if parent_trials_node is not None:
        adapter = node.item['experiment'].refers['adapter']
        for parent in parent_trials_node.root:
            parent.item['trials'] = adapter.forward(parent.item['trials'])

    return node.item, parent_trials_node


def _adapt_children_trials(node, children_trials_nodes):
    """Adapt trials from the children recursively

    .. note::

        To call with node.map(fct, node.children) to connect with children

    """
    for child in children_trials_nodes:
        adapter = child.item['experiment'].refers['adapter']
        for subchild in child:  # Includes child itself
            subchild.item['trials'] = adapter.backward(subchild.item['trials'])

    return node.item, children_trials_nodes


def adapt_trials(trials_tree):
    """Adapt trials recursively so that they are all compatible with current experiment."""
    trials_tree.map(_adapt_parent_trials, trials_tree.parent)
    trials_tree.map(_adapt_children_trials, trials_tree.children)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures for functional serving tests"""
from falcon import testing
import pytest

from orion.serving.webapi import WebApi
from orion.testing import OrionState


@pytest.fixture()
def client():
    """Mock the falcon.API instance for testing with an in memory database"""
    storage = {'type': 'legacy', 'database': {'type': 'EphemeralDB'}}
    with OrionState(storage=storage):
        yield testing.TestClient(WebApi({'storage': storage}))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests for the parsing of the different commands."""
import argparse
import os

import pytest

from orion.core.cli import insert


def _create_parser(need_subparser=True):
    parser = argparse.ArgumentParser()

    if need_subparser:
        subparsers = parser.add_subparsers()
        return parser, subparsers

    return parser


@pytest.mark.usefixtures("clean_db")
def test_insert_command_full_parsing(database, monkeypatch):
    """Test the parsing of all the options of insert"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    parser, subparsers = _create_parser()
    args_list = ["insert", "-n", "test", "--config",
                 "./orion_config_random.yaml", "./black_box.py", "-x=1"]

    insert.add_subparser(subparsers)
    subparsers.choices['insert'].set_defaults(func='')

    args = vars(parser.parse_args(args_list))
    assert args['name'] == 'test'
    assert args['config'].name == './orion_config_random.yaml'
    assert args['user_args'] == ['./black_box.py', '-x=1']



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.format_terminal` -- Utility functions for formatting prints to terminal
==============================================================================================

.. module:: format_terminal
   :platform: Unix
   :synopsis: Functions to build strings for terminal prints

"""


INFO_TEMPLATE = """\
{identification}

{commandline}

{configuration}

{algorithm}

{space}

{metadata}

{refers}

{stats}
"""


def format_info(experiment):
    """Render a string for all info of experiment"""
    info_string = INFO_TEMPLATE.format(
        identification=format_identification(experiment),
        commandline=format_commandline(experiment),
        configuration=format_config(experiment),
        algorithm=format_algorithm(experiment),
        space=format_space(experiment),
        metadata=format_metadata(experiment),
        refers=format_refers(experiment),
        stats=format_stats(experiment))

    return info_string


TITLE_TEMPLATE = """\
{title}
{empty:=<{title_len}}\
"""


def format_title(title):
    """Render a title above an horizontal bar"""
    title_string = TITLE_TEMPLATE.format(
        title=title,
        title_len=len(title),
        empty='')

    return title_string


DICT_EMPTY_LEAF_TEMPLATE = "{tab}{key}\n"
DICT_LEAF_TEMPLATE = "{tab}{key}: {value}\n"
DICT_NODE_TEMPLATE = "{tab}{key}:\n{value}\n"


def format_dict(dictionary, depth=0, width=4, templates=None):
    r"""Render a dict on multiple lines

    Parameters
    ----------
    dictionary: dict
        The dictionary to render
    depth: int
        Tab added at the beginning of every lines
    width: int
        Size of the tab added to each line, multiplied
        by the depth of the object in the dict of dicts.
    templates: dict
        Templates for `empty_leaf`, `leaf` and `dict_node`.
        Default is
        `empty_leaf="{tab}{key}"`
        `leaf="{tab}{key}: {value}\n"`
        `dict_node="{tab}{key}:\n{value}\n"`

    Examples
    --------
    >>> print(format_dict({1: {2: 3, 3: 4}, 2: {3: 4, 4: {5: 6}}}))
    1:
        2: 3
        3: 4
    2:
        3: 4
        4:
            5: 6
    >>> templates = {'leaf': '{tab}{key}={value}\n', 'dict_node': '{tab}{key}:\n{value}\n'}
    >>> print(format_dict({1: {2: 3, 3: 4}, 2: {3: 4, 4: {5: 6}}}, templates=templates))
    1:
        2=3
        3=4
    2:
        3=4
        4:
            5=6

    """
    if isinstance(dictionary, (list, tuple)):
        return format_list(dictionary, depth, width=width, templates=templates)

    # To avoid using mutable objects as default values in function signature.
    if templates is None:
        templates = dict()

    empty_leaf_template = templates.get('empty_leaf', DICT_EMPTY_LEAF_TEMPLATE)
    leaf_template = templates.get('leaf', DICT_LEAF_TEMPLATE)
    node_template = templates.get('dict_node', DICT_NODE_TEMPLATE)

    dict_string = ""
    for key in sorted(dictionary.keys()):
        tab = (" " * (depth * width))
        value = dictionary[key]
        if isinstance(value, (dict, list, tuple)):
            if not value:
                dict_string += empty_leaf_template.format(tab=tab, key=key)
            else:
                subdict_string = format_dict(
                    value, depth + 1, width=width, templates=templates)
                dict_string += node_template.format(tab=tab, key=key, value=subdict_string)
        else:
            dict_string += leaf_template.format(tab=tab, key=key, value=value)

    return dict_string.replace(' \n', '\n').rstrip("\n")


LIST_TEMPLATE = """\
{tab}[
{items}
{tab}]\
"""
LIST_ITEM_TEMPLATE = "{tab}{item}\n"
LIST_NODE_TEMPLATE = "{item}\n"


def format_list(a_list, depth=0, width=4, templates=None):
    r"""Render a list on multiple lines

    Parameters
    ----------
    a_list: list
        The list to render
    depth: int
        Tab added at the beginning of every lines
    width: int
        Size of the tab added to each line, multiplied
        by the depth of the object in the list of lists.
    templates: dict
        Templates for `list`, `item` and `list_node`.
        Default is
        `list="{tab}[\n{items}\n{tab}]"`
        `item="{tab}{item}\n"`
        `list_node="{item}\n"`

    Examples
    --------
    >>> print(format_list([1, [2, 3], 4, [5, 6, 7, 8]]))
    [
        1
        [
            2
            3
        ]
        4
        [
            5
            6
            7
            8
        ]
    ]
    >>> templates = {}
    >>> templates['list'] = '{tab}\n{items}\n{tab}'
    >>> templates['item'] = '{tab}- {item}\n'
    >>> templates['list_node'] = '{tab}{item}\n'
    >>> print(format_list([1, [2, 3], 4, [5, 6, 7, 8]], width=2, templates=templates))
      - 1

        - 2
        - 3

      - 4

        - 5
        - 6
        - 7
        - 8

    """
    # To avoid using mutable objects as default values in function signature.
    if templates is None:
        templates = dict()

    list_template = templates.get('list', LIST_TEMPLATE)
    item_template = templates.get('item', LIST_ITEM_TEMPLATE)
    node_template = templates.get('list_node', LIST_NODE_TEMPLATE)

    tab = (" " * (depth * width))
    list_string = ""
    for i, item in enumerate(a_list, 1):
        subtab = (" " * ((depth + 1) * width))
        if isinstance(item, (dict, list, tuple)):
            item_string = format_dict(item, depth + 1, width=width, templates=templates)
            list_string += node_template.format(tab=subtab, id=i, item=item_string)
        else:
            list_string += item_template.format(tab=subtab, id=i, item=item)

    return list_template.format(tab=tab, items=list_string.rstrip("\n"))


ID_TEMPLATE = """\
{title}
name: {name}
version: {version}
user: {user}
"""


def format_identification(experiment):
    """Render a string for identification section"""
    identification_string = ID_TEMPLATE.format(
        title=format_title("Identification"),
        name=experiment.name,
        version=experiment.version,
        user=experiment.metadata['user'])

    return identification_string


COMMANDLINE_TEMPLATE = """\
{title}
{commandline}
"""


def format_commandline(experiment):
    """Render a string for commandline section"""
    if 'user_args' not in experiment.metadata:
        return ''

    commandline_string = COMMANDLINE_TEMPLATE.format(
        title=format_title("Commandline"),
        commandline=" ".join(experiment.metadata['user_args']))

    return commandline_string


CONFIG_TEMPLATE = """\
{title}
pool size: {experiment.pool_size}
max trials: {experiment.max_trials}
working dir: {experiment.working_dir}
"""


def format_config(experiment):
    """Render a string for config section"""
    config_string = CONFIG_TEMPLATE.format(
        title=format_title("Config"),
        experiment=experiment)

    return config_string


ALGORITHM_TEMPLATE = """\
{title}
{configuration}
"""


def format_algorithm(experiment):
    """Render a string for algorithm section"""
    algorithm_string = ALGORITHM_TEMPLATE.format(
        title=format_title("Algorithm"),
        configuration=format_dict(experiment.configuration['algorithms']))

    return algorithm_string


SPACE_TEMPLATE = """\
{title}
{params}
"""


def format_space(experiment):
    """Render a string for space section"""
    space_string = SPACE_TEMPLATE.format(
        title=format_title("Space"),
        params="\n".join(name + ": " + experiment.space[name].get_prior_string()
                         for name in experiment.space.keys()))

    return space_string


METADATA_TEMPLATE = """\
{title}
user: {experiment.metadata[user]}
datetime: {experiment.metadata[datetime]}
orion version: {experiment.metadata[orion_version]}
VCS:
{vcs}
"""


def format_metadata(experiment):
    """Render a string for metadata section"""
    metadata_string = METADATA_TEMPLATE.format(
        title=format_title("Meta-data"),
        experiment=experiment,
        vcs=format_dict(experiment.metadata.get('VCS', {}), depth=1, width=2))

    return metadata_string


REFERS_TEMPLATE = """\
{title}
root: {root}
parent: {parent}
adapter: {adapter}
"""


def format_refers(experiment):
    """Render a string for refers section"""
    if experiment.node.root is experiment.node:
        root = ''
        parent = ''
        adapter = ''
    else:
        root = experiment.node.root.name
        parent = experiment.node.parent.name
        adapter = "\n" + format_dict(experiment.refers['adapter'].configuration, depth=1, width=2)

    refers_string = REFERS_TEMPLATE.format(
        title=format_title("Parent experiment"),
        root=root,
        parent=parent,
        adapter=adapter)

    return refers_string


STATS_TEMPLATE = """\
{title}
completed: {is_done}
trials completed: {stats[trials_completed]}
best trial:
  id: {stats[best_trials_id]}
  evaluation: {stats[best_evaluation]}
  params:
{best_params}
start time: {stats[start_time]}
finish time: {stats[finish_time]}
duration: {stats[duration]}
"""


NO_STATS_TEMPLATE = """\
{title}
No trials executed...
"""


def format_stats(experiment):
    """Render a string for stat section

    Parameters
    ----------
    experiment: `orion.core.worker.experiment.Experiment`
    templates: dict
        templates for the title and `stats`.
        See `format_title` for more info.

    """
    stats = experiment.stats
    if not stats:
        return NO_STATS_TEMPLATE.format(
            title=format_title("Stats"))

    best_params = get_trial_params(stats['best_trials_id'], experiment)

    stats_string = STATS_TEMPLATE.format(
        title=format_title("Stats"),
        stats=stats,
        best_params=format_dict(best_params, depth=2, width=2),
        is_done=experiment.is_done)

    return stats_string


def get_trial_params(trial_id, experiment):
    """Get params from trial_id in given experiment"""
    best_trial = experiment.get_trial(uid=trial_id)
    if not best_trial:
        return {}

    return best_trial.params



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.experiment`."""

import copy
import datetime
import json
import logging
import tempfile

import pytest

import orion.core
from orion.core.io.space_builder import SpaceBuilder
import orion.core.utils.backward as backward
import orion.core.worker.experiment
from orion.core.worker.experiment import Experiment, ExperimentView
from orion.core.worker.primary_algo import PrimaryAlgo
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage
from orion.testing import OrionState


@pytest.fixture()
def new_config(random_dt):
    """Create a configuration that will not hit the database."""
    new_config = dict(
        name='supernaekei',
        # refers is missing on purpose
        metadata={'user': 'tsirif',
                  'orion_version': 0.1,
                  'user_script': 'abs_path/to_yoyoy.py',
                  'user_config': 'abs_path/hereitis.yaml',
                  'user_args': ['--mini-batch~uniform(32, 256, discrete=True)'],
                  'VCS': {"type": "git",
                          "is_dirty": False,
                          "HEAD_sha": "fsa7df7a8sdf7a8s7",
                          "active_branch": None,
                          "diff_sha": "diff"}},
        version=1,
        pool_size=10,
        max_trials=1000,
        working_dir=None,
        algorithms={'dumbalgo': {}},
        producer={'strategy': 'NoParallelStrategy'},
        # attrs starting with '_' also
        # _id='fasdfasfa',
        # and in general anything which is not in Experiment's slots
        something_to_be_ignored='asdfa'
        )

    backward.populate_space(new_config)

    return new_config


@pytest.fixture
def parent_version_config():
    """Return a configuration for an experiment."""
    config = dict(
        _id='parent_config',
        name="old_experiment",
        version=1,
        algorithms='random',
        metadata={'user': 'corneauf', 'datetime': datetime.datetime.utcnow(),
                  'user_args': ['--x~normal(0,1)']})

    backward.populate_space(config)

    return config


@pytest.fixture
def child_version_config(parent_version_config):
    """Return a configuration for an experiment."""
    config = copy.deepcopy(parent_version_config)
    config['_id'] = 'child_config'
    config['version'] = 2
    config['refers'] = {'parent_id': 'parent_config'}
    config['metadata']['datetime'] = datetime.datetime.utcnow()
    config['metadata']['user_args'].append('--y~+normal(0,1)')
    backward.populate_space(config)
    return config


def _generate(obj, *args, value):
    if obj is None:
        return None

    obj = copy.deepcopy(obj)
    data = obj

    for arg in args[:-1]:
        data = data[arg]

    data[args[-1]] = value
    return obj


base_trial = {
    'experiment': 0,
    'status': 'new',  # new, reserved, suspended, completed, broken
    'worker': None,
    'submit_time': '2017-11-23T02:00:00',
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [
        {'name': 'loss',
         'type': 'objective',  # objective, constraint
         'value': 2}
    ],
    'params': [
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'},
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'lstm_with_attention'}
    ]
}


def generate_trials(status):
    """Generate Trials with different configurations"""
    new_trials = [_generate(base_trial, 'status', value=s) for s in status]

    for i, trial in enumerate(new_trials):
        if trial['status'] != 'new':
            trial['start_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)

    for i, trial in enumerate(new_trials):
        if trial['status'] == 'completed':
            trial['end_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)

    # make each trial unique
    for i, trial in enumerate(new_trials):
        trial['results'][0]['value'] = i
        trial['params'].append({
            'name': '/index',
            'type': 'categorical',
            'value': i
        })

    return new_trials


def assert_protocol(exp, create_db_instance):
    """Transitional method to move away from mongodb"""
    assert exp._storage._db is create_db_instance


def count_experiment(exp):
    """Transitional method to move away from mongodb"""
    return exp._storage._db.count("experiments")


def get_db_from_view(exp):
    """Transitional method to move away from mongodb"""
    return exp._storage._db._db


@pytest.fixture()
def space():
    """Build a space object"""
    return SpaceBuilder().build({'x': 'uniform(0, 10)'})


@pytest.fixture()
def algorithm(space):
    """Build a dumb algo object"""
    return PrimaryAlgo(space, 'dumbalgo')


class TestReserveTrial(object):
    """Calls to interface `Experiment.reserve_trial`."""

    @pytest.mark.usefixtures("create_db_instance")
    def test_reserve_none(self):
        """Find nothing, return None."""
        with OrionState(experiments=[], trials=[]):
            exp = Experiment('supernaekei')
            trial = exp.reserve_trial()
            assert trial is None

    def test_reserve_success(self, random_dt):
        """Successfully find new trials in db and reserve the first one"""
        storage_config = {'type': 'legacy', 'database': {'type': 'EphemeralDB'}}
        with OrionState(trials=generate_trials(['new', 'reserved']),
                        storage=storage_config) as cfg:
            exp = Experiment('supernaekei')
            exp._id = cfg.trials[0]['experiment']

            trial = exp.reserve_trial()

            # Trials are sorted according to hash and 'new' gets position second
            cfg.trials[1]['status'] = 'reserved'
            cfg.trials[1]['start_time'] = random_dt
            cfg.trials[1]['heartbeat'] = random_dt

            assert trial.to_dict() == cfg.trials[1]

    def test_reserve_when_exhausted(self):
        """Return None once all the trials have been allocated"""
        stati = ['new', 'reserved', 'interrupted', 'completed', 'broken']
        with OrionState(trials=generate_trials(stati)) as cfg:
            exp = Experiment('supernaekei')
            exp._id = cfg.trials[0]['experiment']
            assert exp.reserve_trial() is not None
            assert exp.reserve_trial() is not None
            assert exp.reserve_trial() is None

    def test_fix_lost_trials(self):
        """Test that a running trial with an old heartbeat is set to interrupted."""
        trial = copy.deepcopy(base_trial)
        trial['status'] = 'reserved'
        trial['heartbeat'] = datetime.datetime.utcnow() - datetime.timedelta(seconds=60 * 10)
        with OrionState(trials=[trial]) as cfg:
            exp = Experiment('supernaekei')
            exp._id = cfg.trials[0]['experiment']

            assert len(exp.fetch_trials_by_status('reserved')) == 1
            exp.fix_lost_trials()
            assert len(exp.fetch_trials_by_status('reserved')) == 0

    def test_fix_only_lost_trials(self):
        """Test that an old trial is set to interrupted but not a recent one."""
        lost_trial, running_trial = generate_trials(['reserved'] * 2)
        lost_trial['heartbeat'] = datetime.datetime.utcnow() - datetime.timedelta(seconds=60 * 10)
        running_trial['heartbeat'] = datetime.datetime.utcnow()

        with OrionState(trials=[lost_trial, running_trial]) as cfg:
            exp = Experiment('supernaekei')
            exp._id = cfg.trials[0]['experiment']

            assert len(exp.fetch_trials_by_status('reserved')) == 2

            exp.fix_lost_trials()

            reserved_trials = exp.fetch_trials_by_status('reserved')
            assert len(reserved_trials) == 1
            assert reserved_trials[0].to_dict()['params'] == running_trial['params']

            failedover_trials = exp.fetch_trials_by_status('interrupted')
            assert len(failedover_trials) == 1
            assert failedover_trials[0].to_dict()['params'] == lost_trial['params']

    def test_fix_lost_trials_race_condition(self, monkeypatch, caplog):
        """Test that a lost trial fixed by a concurrent process does not cause error."""
        trial = copy.deepcopy(base_trial)
        trial['status'] = 'interrupted'
        trial['heartbeat'] = datetime.datetime.utcnow() - datetime.timedelta(seconds=60 * 10)
        with OrionState(trials=[trial]) as cfg:
            exp = Experiment('supernaekei')
            exp._id = cfg.trials[0]['experiment']

            assert len(exp.fetch_trials_by_status('interrupted')) == 1

            assert len(exp._storage.fetch_lost_trials(exp)) == 0

            def fetch_lost_trials(self, query):
                trial_object = Trial(**trial)
                trial_object.status = 'reserved'
                return [trial_object]

            # Force the fetch of a trial marked as reserved (and lost) while actually interrupted
            # (as if already failed-over by another process).
            with monkeypatch.context() as m:
                m.setattr(exp._storage.__class__, 'fetch_lost_trials', fetch_lost_trials)

                assert len(exp._storage.fetch_lost_trials(exp)) == 1

                with caplog.at_level(logging.DEBUG):
                    exp.fix_lost_trials()

            assert caplog.records[-1].levelname == 'DEBUG'
            assert caplog.records[-1].msg == 'failed'
            assert len(exp.fetch_trials_by_status('interrupted')) == 1
            assert len(exp.fetch_trials_by_status('reserved')) == 0

    def test_fix_lost_trials_configurable_hb(self):
        """Test that heartbeat is correctly being configured."""
        trial = copy.deepcopy(base_trial)
        trial['status'] = 'reserved'
        trial['heartbeat'] = datetime.datetime.utcnow() - datetime.timedelta(seconds=60 * 2)
        with OrionState(trials=[trial]) as cfg:
            exp = Experiment('supernaekei')
            exp._id = cfg.trials[0]['experiment']

            assert len(exp.fetch_trials_by_status('reserved')) == 1

            orion.core.config.worker.heartbeat = 60 * 2

            exp.fix_lost_trials()

            assert len(exp.fetch_trials_by_status('reserved')) == 1

            orion.core.config.worker.heartbeat = 60 * 2 / 10.

            exp.fix_lost_trials()

            assert len(exp.fetch_trials_by_status('reserved')) == 0


def test_update_completed_trial(random_dt):
    """Successfully push a completed trial into database."""
    with OrionState(trials=generate_trials(['new'])) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        trial = exp.reserve_trial()

        results_file = tempfile.NamedTemporaryFile(
            mode='w', prefix='results_', suffix='.log', dir='.', delete=True
        )

        # Generate fake result
        with open(results_file.name, 'w') as file:
            json.dump([{
                'name': 'loss',
                'type': 'objective',
                'value': 2}],
                file
            )
        # --

        exp.update_completed_trial(trial, results_file=results_file)

        yo = get_storage().fetch_trials(exp)[0].to_dict()

        assert len(yo['results']) == len(trial.results)
        assert yo['results'][0] == trial.results[0].to_dict()
        assert yo['status'] == 'completed'
        assert yo['end_time'] == random_dt

        results_file.close()


@pytest.mark.usefixtures("with_user_tsirif")
def test_register_trials(random_dt):
    """Register a list of newly proposed trials/parameters."""
    with OrionState():
        exp = Experiment('supernaekei')
        exp._id = 0

        trials = [
            Trial(params=[{'name': 'a', 'type': 'integer', 'value': 5}]),
            Trial(params=[{'name': 'b', 'type': 'integer', 'value': 6}]),
            ]
        for trial in trials:
            exp.register_trial(trial)

        yo = list(map(lambda trial: trial.to_dict(), get_storage().fetch_trials(exp)))
        assert len(yo) == len(trials)
        assert yo[0]['params'] == list(map(lambda x: x.to_dict(), trials[0]._params))
        assert yo[1]['params'] == list(map(lambda x: x.to_dict(), trials[1]._params))
        assert yo[0]['status'] == 'new'
        assert yo[1]['status'] == 'new'
        assert yo[0]['submit_time'] == random_dt
        assert yo[1]['submit_time'] == random_dt


def test_fetch_all_trials():
    """Fetch a list of all trials"""
    with OrionState(trials=generate_trials(['new', 'reserved', 'completed'])) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        trials = list(map(lambda trial: trial.to_dict(), exp.fetch_trials({})))
        assert trials == cfg.trials


def test_fetch_non_completed_trials():
    """Fetch a list of the trials that are not completed

    trials.status in ['new', 'interrupted', 'suspended', 'broken']
    """
    non_completed_stati = ['new', 'interrupted', 'suspended', 'reserved']
    stati = non_completed_stati + ['completed']
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        trials = exp.fetch_noncompleted_trials()
        assert len(trials) == 4
        assert set(trial.status for trial in trials) == set(non_completed_stati)


def test_is_done_property_with_pending(algorithm):
    """Check experiment stopping conditions when there is pending trials."""
    completed = ['completed'] * 10
    reserved = ['reserved'] * 5
    with OrionState(trials=generate_trials(completed + reserved)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        exp.algorithms = algorithm
        exp.max_trials = 10

        assert exp.is_done

        exp.max_trials = 15

        # There is only 10 completed trials
        assert not exp.is_done

        exp.algorithms.algorithm.done = True

        # Algorithm is done but 5 trials are pending
        assert not exp.is_done


def test_is_done_property_no_pending(algorithm):
    """Check experiment stopping conditions when there is no pending trials."""
    completed = ['completed'] * 10
    broken = ['broken'] * 5
    with OrionState(trials=generate_trials(completed + broken)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        exp.algorithms = algorithm

        exp.max_trials = 15

        # There is only 10 completed trials and algo not done.
        assert not exp.is_done

        exp.algorithms.algorithm.done = True

        # Algorithm is done and no pending trials
        assert exp.is_done


def test_broken_property():
    """Check experiment stopping conditions for maximum number of broken."""
    MAX_BROKEN = 3
    orion.core.config.worker.max_broken = MAX_BROKEN

    stati = (['reserved'] * 10) + (['broken'] * (MAX_BROKEN - 1))
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        assert not exp.is_broken

    stati = (['reserved'] * 10) + (['broken'] * (MAX_BROKEN))
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        assert exp.is_broken


def test_configurable_broken_property():
    """Check if max_broken changes after configuration."""
    MAX_BROKEN = 3
    orion.core.config.worker.max_broken = MAX_BROKEN

    stati = (['reserved'] * 10) + (['broken'] * (MAX_BROKEN))
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']

        assert exp.is_broken

        orion.core.config.worker.max_broken += 1

        assert not exp.is_broken


def test_experiment_stats():
    """Check that property stats is returning a proper summary of experiment's results."""
    NUM_COMPLETED = 3
    stati = (['completed'] * NUM_COMPLETED) + (['reserved'] * 2)
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']
        exp.metadata = {'datetime': datetime.datetime.utcnow()}
        stats = exp.stats
        assert stats['trials_completed'] == NUM_COMPLETED
        assert stats['best_trials_id'] == cfg.trials[3]['_id']
        assert stats['best_evaluation'] == 0
        assert stats['start_time'] == exp.metadata['datetime']
        assert stats['finish_time'] == cfg.trials[0]['end_time']
        assert stats['duration'] == stats['finish_time'] - stats['start_time']
        assert len(stats) == 6


def test_fetch_completed_trials_from_view():
    """Fetch a list of the unseen yet completed trials."""
    non_completed_stati = ['new', 'interrupted', 'suspended', 'reserved']
    stati = non_completed_stati + ['completed']
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']
        exp_view = ExperimentView(exp)

        trials = exp_view.fetch_trials_by_status('completed')
        assert len(trials) == 1
        assert trials[0].status == 'completed'


def test_view_is_done_property_with_pending(algorithm):
    """Check experiment stopping conditions from view when there is pending trials."""
    completed = ['completed'] * 10
    reserved = ['reserved'] * 5
    with OrionState(trials=generate_trials(completed + reserved)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']
        exp.algorithms = algorithm
        exp.max_trials = 10

        exp_view = ExperimentView(exp)

        assert exp_view.is_done

        exp.max_trials = 15

        # There is only 10 completed trials
        assert not exp_view.is_done

        exp.algorithms.algorithm.done = True

        # Algorithm is done but 5 trials are pending
        assert not exp_view.is_done


def test_view_is_done_property_no_pending(algorithm):
    """Check experiment stopping conditions from view when there is no pending trials."""
    completed = ['completed'] * 10
    broken = ['broken'] * 5
    with OrionState(trials=generate_trials(completed + broken)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']
        exp.algorithms = algorithm
        exp.max_trials = 100

        exp_view = ExperimentView(exp)

        exp.algorithms = algorithm

        exp.max_trials = 15

        # There is only 10 completed trials and algo not done.
        assert not exp_view.is_done

        exp.algorithms.algorithm.done = True

        # Algorithm is done and no pending trials
        assert exp_view.is_done


def test_experiment_view_stats():
    """Check that property stats from view is consistent."""
    NUM_COMPLETED = 3
    stati = (['completed'] * NUM_COMPLETED) + (['reserved'] * 2)
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment('supernaekei')
        exp._id = cfg.trials[0]['experiment']
        exp.metadata = {'datetime': datetime.datetime.utcnow()}

        exp_view = ExperimentView(exp)

        stats = exp_view.stats
        assert stats['trials_completed'] == NUM_COMPLETED
        assert stats['best_trials_id'] == cfg.trials[3]['_id']
        assert stats['best_evaluation'] == 0
        assert stats['start_time'] == exp_view.metadata['datetime']
        assert stats['finish_time'] == cfg.trials[0]['end_time']
        assert stats['duration'] == stats['finish_time'] - stats['start_time']
        assert len(stats) == 6


def test_experiment_view_protocol_read_only():
    """Verify that wrapper experiments' protocol is read-only"""
    with OrionState():
        exp = Experiment('supernaekei')

        exp_view = ExperimentView(exp)

        # Test that _protocol.set_trial_status indeed exists
        exp_view._experiment._storage._storage.set_trial_status

        with pytest.raises(AttributeError):
            exp_view._experiment._storage.set_trial_status


def test_view_is_broken():
    """Tests that property ``is_broken`` is accessible in an experiment's view"""
    broken = ['broken'] * 5
    with OrionState(trials=generate_trials(broken)) as cfg:
        exp = Experiment('test-experiment')
        exp._id = cfg.trials[0]['experiment']

        assert exp.is_broken

        exp_view = ExperimentView(exp)
        assert exp_view.is_broken



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils` -- Package-wide useful routines
=======================================================

.. module:: utils
   :platform: Unix
   :synopsis: Helper functions useful in possibly all :mod:`orion.core`'s modules.
"""

from abc import ABCMeta
from collections import defaultdict
from glob import glob
from importlib import import_module
import logging
import os

import pkg_resources

log = logging.getLogger(__name__)


def nesteddict():
    """
    Define type of arbitrary nested defaultdicts
    Extend defaultdict to arbitrary nested levels.
    """
    return defaultdict(nesteddict)


class Factory(ABCMeta):
    """Instantiate appropriate wrapper for the infrastructure based on input
    argument, ``of_type``.

    Attributes
    ----------
    types : list of subclasses of ``cls.__base__``
       Updated to contain all possible implementations currently. Check out code.
    typenames : list of str
       Names of implemented wrapper classes, correspond to possible ``of_type``
       values.

    """

    def __init__(cls, names, bases, dictionary):
        """Search in directory for attribute names subclassing `bases[0]`"""
        super(Factory, cls).__init__(names, bases, dictionary)

        cls.modules = []
        base = import_module(cls.__base__.__module__)
        try:
            py_files = glob(os.path.abspath(os.path.join(base.__path__[0], '[A-Za-z]*.py')))
            py_mods = map(lambda x: '.' + os.path.split(os.path.splitext(x)[0])[1], py_files)
            for py_mod in py_mods:
                cls.modules.append(import_module(py_mod, package=cls.__base__.__module__))
        except AttributeError:
            # This means that base class and implementations reside in a module
            # itself and not a subpackage.
            pass

        # Get types advertised through entry points!
        for entry_point in pkg_resources.iter_entry_points(cls.__name__):
            entry_point.load()
            log.debug("Found a %s %s from distribution: %s=%s",
                      entry_point.name, cls.__name__,
                      entry_point.dist.project_name, entry_point.dist.version)

        # Get types visible from base module or package, but internal
        def get_all_subclasses(parent):
            """Get set of subclasses recursively"""
            subclasses = set()
            for subclass in parent.__subclasses__():
                subclasses.add(subclass)
                subclasses |= get_all_subclasses(subclass)

            return subclasses

        cls.types = list(get_all_subclasses(cls.__base__))
        cls.types = [class_ for class_ in cls.types if class_.__name__ != cls.__name__]
        cls.typenames = list(map(lambda x: x.__name__.lower(), cls.types))
        log.debug("Implementations found: %s", cls.typenames)

    def __call__(cls, of_type, *args, **kwargs):
        """Create an object, instance of ``cls.__base__``, on first call.

        :param of_type: Name of class, subclass of ``cls.__base__``, wrapper
           of a database framework that will be instantiated on the first call.
        :param args: positional arguments to initialize ``cls.__base__``'s instance (if any)
        :param kwargs: keyword arguments to initialize ``cls.__base__``'s instance (if any)

        .. seealso::
           `Factory.typenames` for values of argument `of_type`.

        .. seealso::
           Attributes of ``cls.__base__`` and ``cls.__base__.__init__`` for
           values of `args` and `kwargs`.

        .. note:: New object is saved as `Factory`'s internal state.

        :return: The object which was created on the first call.
        """
        for inherited_class in cls.types:
            if inherited_class.__name__.lower() == of_type.lower():
                return inherited_class.__call__(*args, **kwargs)

        error = "Could not find implementation of {0}, type = '{1}'".format(
            cls.__base__.__name__, of_type)
        error += "\nCurrently, there is an implementation for types:\n"
        error += str(cls.typenames)
        raise NotImplementedError(error)
